#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Routines to convert Phoenix data to day-long mseed files.

Maeva Pourpoint - IRIS/PASSCAL
"""

from __future__ import annotations

from obspy import read, Stream, Trace, UTCDateTime
from obspy.core.trace import Stats
from pandas import Series
from pathlib import Path
from typing import List, Tuple, TYPE_CHECKING

from phoenix2seed.logging import parse_config_ini, setup_logger

if TYPE_CHECKING:
    from phoenix2seed.phoenix_data import PhoenixData

# Read config.ini file and set up logging
config = parse_config_ini()
logger = setup_logger(__name__)

DAY_S = 86400  # s
PRECISION = 0.000001  # s


def init_stats(phoenix_data: PhoenixData) -> Stats:
    """Instantiate Stats object with proper network code and station name."""
    stats = Stats()
    stats.network = phoenix_data.stats["net"]
    stats.station = phoenix_data.stats["sta"]
    return stats


def update_stats(datum: Series, stats: Stats) -> Stats:
    """
    Update Stats object with proper channel name, location code, sample rate
    and start time of the trace.
    """
    stats.channel = datum["cha_name"]
    stats.location = datum["loc"]
    stats.sampling_rate = datum["sample_rate"]
    stats.starttime = datum["starttime"]
    return stats


def mseed_startend_time(trace: Trace) -> Tuple[UTCDateTime, UTCDateTime,
                                               UTCDateTime, UTCDateTime]:
    """
    Reads in a trace and returns the start and end time for the current
    and next days.
    """
    current_start = trace.stats.starttime
    current_end = trace.stats.endtime
    next_start = UTCDateTime(current_start.year, current_start.month,
                             current_start.day) + DAY_S
    if current_end > next_start:
        next_end = current_end
        current_end = next_start - PRECISION
    else:
        next_end = None
    return current_start, current_end, next_start, next_end


def trim_traces(st_all: Stream) -> List:
    """
    Trim traces if their start time and end time don't fall on the same day.
    Necessary step to create day-long mseed.
    """
    traces = []
    for tr in st_all:
        while True:
            current_start, current_end, next_start, next_end = mseed_startend_time(tr)
            tr_day = tr.copy()
            # If next_end is not None, trim trace into two traces:
            # - current day trace
            # - next day trace
            if next_end:
                tr_day.trim(starttime=current_start, endtime=current_end, nearest_sample=False)
                tr_nextday = tr.copy()
                tr_nextday.trim(starttime=next_start, endtime=next_end, nearest_sample=False)
                # Next day trace may need to be trimmed as well.
                tr = tr_nextday
            traces.append(tr_day)
            if next_end is None:
                break
    return traces


def get_daylong_streams(traces: List) -> List:
    """
    Create list of Stream objects.
    Each Stream containing a list of same-day traces.
    """
    st_days = []
    st_day = Stream()
    if len(traces) == 1:
        st_day += traces[0]
        st_days.append(st_day)
    else:
        for tr1, tr2 in zip(traces[:-1], traces[1:]):
            st_day += tr1
            if tr2.stats.starttime.julday > tr1.stats.starttime.julday:
                st_days.append(st_day)
                st_day = Stream()
            if tr2 == traces[-1]:
                st_day += tr2
                st_days.append(st_day)
    return st_days


def handle_existing_mseed_file(st_current: Stream, filename: Path) -> Stream:
    """
    If for a given day a mseed file already exists, read existing mseed file
    into a Stream object and merge that Stream object to the current Stream
    object.
    """
    st_existing = read(str(filename))
    st_new = st_existing + st_current
    return st_new


def write_daylong_mseed(phoenix_data: PhoenixData) -> None:
    """Create daylong mseed files for each channel and location code."""
    path2mseed = phoenix_data.output_mseed.joinpath(phoenix_data.stats["sta"])
    if not path2mseed.is_dir():
        path2mseed.mkdir(parents=True)
    stats = init_stats(phoenix_data)
    chas = set(phoenix_data.data_frame["cha_name"])
    locs = set(phoenix_data.data_frame["loc"])
    for cha in chas:
        logger.info("Writing day-long mseed files for channel: {}.".format(cha))
        for loc in locs:
            # Get indexes of row for which channel name and location code
            # respectively match cha and loc
            ind = phoenix_data.data_frame["cha_name"].eq(cha) & phoenix_data.data_frame["loc"].eq(loc)
            data = phoenix_data.data_frame[ind]
            st_all = []
            # For each filtered row, create a trace and update stats of that trace
            for i in range(len(data)):
                datum = data.iloc[i]
                tr = Trace()
                tr.stats = update_stats(datum, stats)
                tr.data = datum["samples"]
                st_all.append(tr.copy())
            # Trim traces if necessary and create day-long Stream objects
            traces = trim_traces(st_all)
            st_days = get_daylong_streams(traces)
            # For each day-long Stream objects, write day-long mseed file
            for st_day in st_days:
                stats = st_day[0].stats
                filename = path2mseed.joinpath('.'.join([stats.station,
                                                         stats.network, loc, cha,
                                                         str(stats.starttime.year),
                                                         str(stats.starttime.julday).zfill(3)]))
                if filename.is_file():
                    logger.info("The mseed file {} already exists. Merging to "
                                "it. There will be no duplication of existing data."
                                .format(filename))
                    st_day = handle_existing_mseed_file(st_day, filename)
                st_day.merge(-1)
                logger.info("Writing day-long mseed file: {}.".format(filename))
                st_day.write(filename, format='MSEED')
    logger.info("Writing of day-long mseed files successfully completed!")
