# -*- coding: utf-8 -*-
"""
phoenix2seed GUI Framework -  View.

Maeva Pourpoint - IRIS/PASSCAL
"""

from dataclasses import fields
from operator import attrgetter, methodcaller
from pathlib import Path
from PySide6 import QtCore, QtWidgets, QtGui

from phoenix2seed.data_conversion import write_daylong_mseed
from phoenix2seed.gui_model import TreeModel
from phoenix2seed.gui_widgets import (load_ui, BaseWidget, NetWidgets,
                                      StaWidgets, RunWidgets, HelpWidgets,
                                      get_channel_widgets)
from phoenix2seed.logging import parse_config_ini, setup_logger
from phoenix2seed.metadata_category import MSG_MAP
from phoenix2seed.metadata_conversion import write_stationxml
from phoenix2seed.phoenix_metadata import PhoenixMetadata, IDS
from phoenix2seed.utils import get_ids, NUM_CHA_MAX


# Read config.ini file and set up logging
config = parse_config_ini()
logger = setup_logger(__name__)

VALID_NUM_CHAS = {'elec': [0, 2], 'mag': [2, 3]}


class PhoenixWindow(*load_ui('mainwindow.ui')):

    def __init__(self, phoenix_data, phoenix_md, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.setWindowTitle(f"{config['software']['name']} - {config['software']['version']}")
        self.phoenix_data = phoenix_data
        self.phoenix_md = phoenix_md
        # -- Tool bar --
        self.help_window = HelpWidgets()
        self.actionHelp.triggered.connect(self.toggle_help_window)
        self.actionLoad_MD.triggered.connect(self.load_md)
        self.actionSave_MD.triggered.connect(self.save_md)
        # -- Network --
        self.net_widgets = NetWidgets(self)
        self.uiNetEditor.addWidget(self.net_widgets)
        # -- Station --
        self.sta_widgets = StaWidgets(self)
        self.uiStaEditor.addWidget(self.sta_widgets)
        # -- Runs --
        self.num_runs = phoenix_md.num_runs
        self.run_list = phoenix_md.run_list
        self.set_run_editor_forms()
        # -- Electric --
        ElecWidgets = get_channel_widgets([BaseWidget, *load_ui("electricwidget.ui")])
        self.set_cha_editor_forms('elec', ElecWidgets)
        # -- Magnetic --
        MagWidgets = get_channel_widgets([BaseWidget, *load_ui("magneticwidget.ui")])
        self.set_cha_editor_forms('mag', MagWidgets)
        self.uiMagEditor.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignLeft)
        # -- Set Model, Process metadata, Convert data and metadata to archivable format --
        self.set_model()
        self.processing_conversion_steps()

    def toggle_help_window(self):
        if self.help_window.isVisible():
            self.help_window.hide()
        else:
            self.help_window.show()

    def create_message_box(self, text, details, icon):
        """
        Create generic message box. This box will be used to display different
        texts to alert the user to a given situation.
        """
        font = QtGui.QFont()
        font.setFamily("Arial")
        font.setPointSize(13)
        msg_box = QtWidgets.QMessageBox(self)
        msg_box.setFont(font)
        msg_box.setText(text)
        msg_box.setInformativeText(details)
        if icon:
            msg_box.setIcon(icon)
        return msg_box

    def reset_model(self, filename):
        """
        Reset current metadata model.
        Called when loading a previously saved metadata model.
        """
        phoenix_md_saved = PhoenixMetadata.load_md(filename)
        run_list = list(set(self.run_list) & set(phoenix_md_saved.run_list))
        # Set/Reset metadata at network level
        self.phoenix_md.net = phoenix_md_saved.net
        # Set/Reset metadata at station level
        # Reset station name to None to ensure that the user checks preloaded
        # metadata before conversion (per data group suggestion)
        phoenix_md_saved.sta.archive_id = None
        self.phoenix_md.sta = phoenix_md_saved.sta
        # Set/Reset metadata at run level
        # Instantiate Run data class with proper run_id, start and end times if
        # field sheets were not parsed and initialize GUI attributes (only)
        # using previously saved ones.
        if not self.phoenix_md.run:
            self.phoenix_md.run.extend(self.phoenix_md.init_run_props())
        for x in phoenix_md_saved.run:
            if x.run_id not in run_list:
                continue
            ind_run = self.phoenix_md.run_list.index(x.run_id)
            md_props = x.__dict__
            md_gui = [md_field.name for md_field in fields(x) if md_field.metadata.get('gui')]
            for key, val in md_props.items():
                if key in md_gui:
                    setattr(self.phoenix_md.run[ind_run], key, val)
                if key in self.phoenix_md.run[ind_run].md_missing:
                    self.phoenix_md.run[ind_run].md_missing.discard(key)
                if key in self.phoenix_md.run[ind_run].md_invalid:
                    self.phoenix_md.run[ind_run].md_invalid.discard(key)
        # Set/Reset metadata at data channel level
        self.phoenix_md.elec = [x for x in phoenix_md_saved.elec if x.run_id in run_list]
        self.phoenix_md.mag = [x for x in phoenix_md_saved.mag if x.run_id in run_list]
        self.set_model()
        for run_id in run_list:
            self.update_cha_selection_forms(run_id, 'elec')
            self.update_cha_selection_forms(run_id, 'mag')

    def load_md(self, path2savedmd=None):
        """Load previously saved metadata model."""
        if path2savedmd:
            filename = str(path2savedmd)
        else:
            filename, _ = QtWidgets.QFileDialog.getOpenFileName(parent=self,
                                                                caption='Select metadata file to open',
                                                                dir=str(Path.cwd()),
                                                                filter='Metadata file (*.pkl)')
        if filename:
            text = "Loading previously saved metadata..."
            details = ("If you are creating mseed and StationXML files for a "
                       "different station than the one for which you are "
                       "uploading previously saved metadata, please check the "
                       "validity of all your metadata fields before proceeding "
                       "with conversion!")
            icon = QtWidgets.QMessageBox.Information
            msg_box = self.create_message_box(text, details, icon)
            msg_box.exec()
            self.reset_model(filename)

    def save_md(self):
        """Save current metadata model."""
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(parent=self,
                                                            caption='Save inputted metadata',
                                                            dir=str(Path.cwd().joinpath('phoenix_metadata.pkl')),
                                                            filter='Metadata file (*.pkl)')
        if filename:
            self.phoenix_md.save_md(filename)

    def get_run_widgets(self, run_id):
        return getattr(self, f'run_{run_id}_widgets')

    def set_run_widgets(self, run_id, val):
        setattr(self, f'run_{run_id}_widgets', val)

    def set_num_runs_form(self):
        """
        Set up number of runs form. There is one form for all runs. This form
        is non-editable and reports to the user how many runs were recorded.
        """
        layout_num_runs = QtWidgets.QFormLayout()
        uiNumRuns = QtWidgets.QLineEdit()
        uiNumRuns.setText(str(self.num_runs))
        uiNumRuns.setReadOnly(True)
        uiNumRuns.setSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Preferred)
        uiNumRuns.setMinimumWidth(30)
        layout_num_runs.addRow('Number of runs *', uiNumRuns)
        layout_num_runs.setFormAlignment(QtCore.Qt.AlignLeft)
        return layout_num_runs

    def set_run_starttime_form(self, run_id):
        """
        Set up form for the start time of a given run. There is one form per
        run id. This form is non-editable and provides additional background
        information to the user.
        """
        # Get run start time
        starttime = self.phoenix_data.stats['acq_start_utc'][run_id]
        starttime_reformat = starttime.datetime.strftime("%Y-%m-%d %H:%M:%S")
        # Set up layout form
        layout_run_starttime = QtWidgets.QFormLayout()
        uiRunStarttime = QtWidgets.QLineEdit()
        uiRunStarttime.setText(starttime_reformat)
        uiRunStarttime.setReadOnly(True)
        uiRunStarttime.setSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Preferred)
        uiRunStarttime.setMinimumWidth(175)
        layout_run_starttime.addRow('Start time of run *', uiRunStarttime)
        layout_run_starttime.setFormAlignment(QtCore.Qt.AlignLeft)
        return layout_run_starttime

    def submit_run_cha_md(self, run_id, cha_type=None):
        """
        Submit run and channel metadata entries provided by the user up to this
        point before checking validity of metadata fields to be copied.
        Submit policy set in BaseWidget is Manual.
        """
        # -- Run --
        run = self.get_run_widgets(run_id)
        run.mapper.submit()
        # -- Electric or Magnetic channel --
        if cha_type:
            num_chas_submit = getattr(self, f'run_{run_id}_num_{cha_type}', NUM_CHA_MAX[cha_type])
            for ind in range(1, num_chas_submit + 1):
                cha = self.get_cha_widgets(run_id, ind, cha_type)
                cha.mapper.submit()

    def check_before_reset(self, run_id_same_as, run_id, cha_type):
        """
        Check/Validate metadata fields for both runs before copying electric or
        magnetic metadata fields from one run to the other.
        """
        logger.info(f"{'-'*20} Checking metadata fields for run '{run_id_same_as}' "
                    f"before copying {MSG_MAP[cha_type.capitalize()]} metadata "
                    f"fields from run '{run_id_same_as}' to run '{run_id}'. Any "
                    "incomplete or invalid metadata fields may prevent the "
                    "auto-filling process from succesfully completing. Look for "
                    f"potential error messages below {'-'*20}.")
        self.validate_run(run_id_same_as)
        if cha_type != 'run':
            self.validate_cha(cha_type, run_id_same_as)
        logger.info(f"{'-'*20} End of metadata field checks for run "
                    f"'{run_id_same_as}' {'-'*20}.")

    def reset_run_model(self, run_id_same_as, run_id):
        """
        Reset current run metadata model based on a user-selected run id.
        Called when selecting the "Same as run" option for run_id > a under the
        "Runs" tab.
        """
        self.submit_run_cha_md(run_id_same_as)
        self.check_before_reset(run_id_same_as, run_id, 'run')
        md = self.phoenix_md.for_gui
        key = f'Run_{run_id}'
        key_same_as = f'Run_{run_id_same_as}'
        md['Run'][key].update(md['Run'][key_same_as])
        model = TreeModel(("Field", "Value"), md)
        run = self.get_run_widgets(run_id)
        run.set_model(model, f'Run_{run_id}')

    def set_cha_same_as_forms(self, level, run_id):
        """
        Set up "Same as run" form.
        Called when there is more than one run and selecting the "Same as run" option.
        """
        layout_same_as = QtWidgets.QFormLayout()
        uiRunIdSameAs = QtWidgets.QComboBox()
        run_ids = self.run_list[:self.run_list.index(run_id)]
        uiRunIdSameAs.addItems(run_ids)
        cha_type = MSG_MAP[level.capitalize()]
        for ind, id_ in enumerate(run_ids):
            tip_msg = f"Same {cha_type} metadata field values as Run {id_}."
            uiRunIdSameAs.setItemData(ind, tip_msg, QtCore.Qt.ToolTipRole)
        if level == 'run':
            reset_method = 'reset_run_model'
            params = run_id
        else:
            reset_method = 'reset_cha_model'
            params = [run_id, level]
        uiRunIdSameAs.textActivated.connect(lambda text, val=params: methodcaller(reset_method, text, val)(self))
        layout_same_as.addRow('Same as run', uiRunIdSameAs)
        layout_same_as.setFormAlignment(QtCore.Qt.AlignLeft)
        return layout_same_as

    def set_run_editor_forms(self):
        """Set up run editor form. There is one form per run id."""
        layout_num_runs = self.set_num_runs_form()
        self.sta_widgets.uiRunsEditor.addLayout(layout_num_runs)
        self.sta_widgets.uiRunsEditor.addSpacing(5)
        uiTabs = QtWidgets.QTabWidget()
        setattr(self, 'uiRunTabs', uiTabs)
        self.sta_widgets.uiRunsEditor.addWidget(uiTabs)
        for run_id in self.run_list:
            uiRun = QtWidgets.QWidget()
            layout_run_option = QtWidgets.QHBoxLayout()
            layout_run_starttime = self.set_run_starttime_form(run_id)
            layout_run_option.addLayout(layout_run_starttime)
            if run_id != 'a':
                layout_run_same_as = self.set_cha_same_as_forms('run', run_id)
                layout_run_option.addLayout(layout_run_same_as)
            layout_run = QtWidgets.QVBoxLayout()
            layout_run.addLayout(layout_run_option)
            layout_run.addSpacing(5)
            self.set_run_widgets(run_id, RunWidgets(self))
            layout_run.addWidget(self.get_run_widgets(run_id))
            uiRun.setLayout(layout_run)
            setattr(self, f'uiRun_{run_id}', uiRun)
            uiTabs.addTab(uiRun, 'Run ' + run_id)

    def set_cha_widgets(self, run_id, ind, val, cha_type):
        setattr(self, f'run_{run_id}_{cha_type}_{ind}_widgets', val)

    def get_cha_widgets(self, run_id, ind, cha_type):
        return getattr(self, f'run_{run_id}_{cha_type}_{ind}_widgets')

    def update_combo_chas(self, ind, params):
        """
        Update channel port combo box values based on channel port selected
        for other channels.
        Used for both the electric and magnetic channels.
        """
        run_id, cha_type, ind_e_pair = params
        cha_1 = self.get_cha_widgets(run_id, ind_e_pair, cha_type)
        combo_parent = cha_1.uiChaPort
        cha_2 = self.get_cha_widgets(run_id, ind_e_pair + 1, cha_type)
        combo_child = cha_2.uiChaPort
        combo_child.clear()
        chas = combo_parent.itemData(ind)
        if chas is not None:
            for i in range(len(chas)):
                combo_child.addItem(chas[i], [x for x in chas if x != chas[i]])

    def set_cha_forms(self, run_id, cha_type, ChaWidgets):
        """
        Set up the electrode pair or coil form. There is one form per channel
        and that form includes all the  metadata fields required for archival of
        that electric or magnetic channel.
        """
        layout_chas = QtWidgets.QHBoxLayout()
        ports = [f"{IDS[cha_type]['cha_id']}{i+1}" for i in range(NUM_CHA_MAX[cha_type])]
        sensor = IDS[cha_type]['sensor']
        for ind in range(1, NUM_CHA_MAX[cha_type] + 1):
            self.set_cha_widgets(run_id, ind, ChaWidgets(self, cha_type), cha_type)
            cha = self.get_cha_widgets(run_id, ind, cha_type)
            # Set up channel port combo box for electrode pair 1  or coil 1.
            if ind == 1:
                for i in range(NUM_CHA_MAX[cha_type]):
                    cha.uiChaPort.addItem(ports[i], [x for x in ports if x != ports[i]])
            # Set up channel port combo box for electrode pair 2 or coils 2 and
            # 3 based on parent channel port combo boxes (respectively E1 or H1
            # and H2).
            if ind > 1:
                parent = self.get_cha_widgets(run_id, ind-1, cha_type).uiChaPort
                self.update_combo_chas(parent.currentIndex(), [run_id, cha_type, ind-1])
            # Value of the channel port for electrode pair 2 or coil 3 cannot be
            # directly selected by the user. Undirectly updated by updating
            # parent combo boxes.
            if ind < NUM_CHA_MAX[cha_type]:
                cha.uiChaPort.currentIndexChanged.connect(lambda ind_, params=[run_id, cha_type, ind]: self.update_combo_chas(ind_, params))  # noqa: E501
            cha.groupBox.setTitle(' '.join([sensor.capitalize(), str(ind)]))
            layout_chas.addWidget(cha)
        return layout_chas

    def update_cha_forms(self, num_chas, params):
        """
        Disable or enable electrode pair/coil form based on the selected number
        of electrode pairs/ coils.
        """
        run_id, cha_type = params
        for ind in range(1, NUM_CHA_MAX[cha_type] + 1):
            cha = self.get_cha_widgets(run_id, ind, cha_type)
            if ind in range(1, int(num_chas) + 1):
                cha.groupBox.setEnabled(True)
            else:
                cha.groupBox.setEnabled(False)
        setattr(self, f'run_{run_id}_num_{cha_type}', int(num_chas))

    def set_cha_selection_forms(self, run_id, cha_type):
        """
        Set up electrode/coil selection form. There is one form per run id. That
        form allows the user to select how many electrode pairs/coils were
        deployed for that specific run.
        Based on the selected number, the electrode pair/coil form for the
        non-deployed pair(s) will be disabled.
        """
        layout_selection = QtWidgets.QFormLayout()
        uiNumChas = QtWidgets.QComboBox()
        uiNumChas.addItems([str(x) for x in VALID_NUM_CHAS[cha_type]])
        sensor = IDS[cha_type]['sensor']
        for ind, num in enumerate(VALID_NUM_CHAS[cha_type]):
            tip_msg = f'Number of {sensor}s deployed: {num}'
            uiNumChas.setItemData(ind, tip_msg, QtCore.Qt.ToolTipRole)
        uiNumChas.setObjectName(f'run_{run_id}_num_{cha_type}')
        uiNumChas.textActivated.connect(lambda text, params=[run_id, cha_type]: self.update_cha_forms(text, params))
        setattr(self, f'run_{run_id}_num_{cha_type}_widgets', uiNumChas)
        layout_selection.addRow(f'Number of {sensor}s *', uiNumChas)
        layout_selection.setFormAlignment(QtCore.Qt.AlignLeft)
        return layout_selection

    def update_cha_selection_forms(self, run_id, cha_type):
        """
        Update electrode/coil selection form for given run_id if the number of
        electrode pairs/coils can be infered from the 'channels recorded'
        metadata field.
        """
        cha_id = IDS[cha_type]['cha_id']
        chas_rec = self.phoenix_md.get_chas_rec(cha_id, run_id) if self.phoenix_md.run else []
        if chas_rec:
            num_chas = len(chas_rec)
            uiNumChas = getattr(self, f'run_{run_id}_num_{cha_type}_widgets')
            uiNumChas.setCurrentIndex(VALID_NUM_CHAS[cha_type].index(num_chas))
            self.update_cha_forms(num_chas, [run_id, cha_type])

    def reset_cha_model(self, run_id_same_as, params):
        """
        Reset current electric metadata model based on a user-selected run id.
        Called when selecting the "Same as run" option for run_id > a.
        """
        run_id, cha_type = params
        self.submit_run_cha_md(run_id_same_as, cha_type)
        self.check_before_reset(run_id_same_as, run_id, cha_type)
        md = self.phoenix_md.for_gui
        cat = cha_type.capitalize()
        for ind_cha in range(1, NUM_CHA_MAX[cha_type] + 1):
            key = f'Run_{run_id}_{cha_type}_{ind_cha}'
            key_same_as = f'Run_{run_id_same_as}_{cha_type}_{ind_cha}'
            md[cat][key].update(md[cat][key_same_as])
        model = TreeModel(("Field", "Value"), md)
        for ind_cha in range(1, NUM_CHA_MAX[cha_type] + 1):
            cha = self.get_cha_widgets(run_id, ind_cha, cha_type)
            cha.set_model(model, f'Run_{run_id}_{cha_type}_{ind_cha}')

    def set_cha_editor_forms(self, cha_type, ChaWidgets):
        """
        Set up the electric/coil editor form for all runs.
        This form includes the electrode/coil selection form, electrode
        pair/coil form and "same as run" form (if the run id > a).
        """
        uiTabs = QtWidgets.QTabWidget()
        cat = cha_type.capitalize()
        setattr(self, f'uiRunTabs_{cat}', uiTabs)
        editor = getattr(self, f"ui{cat}Editor")
        editor.addWidget(uiTabs)
        for run_id in self.run_list:
            layout_chas = self.set_cha_forms(run_id, cha_type, ChaWidgets)
            layout_option = QtWidgets.QHBoxLayout()
            layout_selection = self.set_cha_selection_forms(run_id, cha_type)
            self.update_cha_selection_forms(run_id, cha_type)
            layout_option.addLayout(layout_selection)
            if run_id != 'a':
                layout_same_as = self.set_cha_same_as_forms(cha_type, run_id)
                layout_option.addLayout(layout_same_as)
            layout = QtWidgets.QVBoxLayout()
            layout.addLayout(layout_option)
            layout.addLayout(layout_chas)
            layout.setAlignment(QtCore.Qt.AlignTop)
            uiRun = QtWidgets.QWidget()
            uiRun.setLayout(layout)
            setattr(self, f'uiRun_{run_id}_{cat}', uiRun)
            uiTabs.addTab(uiRun, 'Run ' + run_id)

    def set_model(self):
        """
        Set tree model. Includes metadata at the net, sta, run, elec and mag
        levels.
        """
        model = TreeModel(("Field", "Value"), self.phoenix_md.for_gui)
        self.net_widgets.set_model(model)
        self.sta_widgets.set_model(model)
        for run_id in self.run_list:
            run = self.get_run_widgets(run_id)
            run.set_model(model, f'Run_{run_id}')
        for cha_type in ['elec', 'mag']:
            for run_id, ind_cha in get_ids(self.num_runs, cha_type):
                cha = self.get_cha_widgets(run_id, ind_cha, cha_type)
                cha.set_model(model, f'Run_{run_id}_{cha_type}_{ind_cha}')

    def flag_tab_bar(self, tabwidget, cat, num_missing_invalid):
        """
        Flag panel name in the main window if metadata fields under that panel
        are either invalid or missing.
        """
        bar = tabwidget.tabBar()
        ind = tabwidget.indexOf(getattr(self, f'ui{cat}'))
        if num_missing_invalid > 0:
            bar.setTabTextColor(ind, QtGui.QColor(255, 0, 0))
        else:
            bar.setTabTextColor(ind, QtGui.QColor(0, 0, 0))

    def flag_net_md(self):
        """
        Flag network metadata fields that are required but unfilled or invalid.
        """
        req_fields_n = self.phoenix_md.net.md_missing | self.phoenix_md.net.md_invalid
        self.net_widgets.set_flags(req_fields_n)
        num_missing_invalid_n = len(req_fields_n)
        self.flag_tab_bar(self.uiMainTabs, 'Net', num_missing_invalid_n)
        return num_missing_invalid_n

    def validate_net(self):
        """Validate network metadata entry."""
        md_fields_n = self.net_widgets.model
        self.phoenix_md.populate_net_props(md_fields_n)
        num_missing_invalid_n = self.flag_net_md()
        return num_missing_invalid_n

    def flag_sta_md(self):
        """Flag station metadata fields that are required but unfilled or invalid."""
        req_fields_s = self.phoenix_md.sta.md_missing | self.phoenix_md.sta.md_invalid
        self.sta_widgets.set_flags(req_fields_s)
        num_missing_invalid_s = len(req_fields_s)
        self.flag_tab_bar(self.uiMainTabs, 'Sta', num_missing_invalid_s)
        return num_missing_invalid_s

    def validate_sta(self):
        """Validate station metadata entry."""
        md_fields_s = self.sta_widgets.model
        reset = self.phoenix_md.populate_sta_props(md_fields_s)
        # If start or end time of the station was reset, reset current sta metadata model
        if reset:
            model = TreeModel(("Field", "Value"), self.phoenix_md.for_gui)
            self.sta_widgets.set_model(model)
        num_missing_invalid_s = self.flag_sta_md()
        return num_missing_invalid_s

    def flag_run_md(self, run_list):
        """Flag run metadata fields that are required but unfilled or invalid."""
        req_fields_r = [run.md_missing | run.md_invalid
                        for run in self.phoenix_md.run
                        if run.resource_id.split(':')[1] in run_list]
        num_missing_invalid_r = sum([len(x) for x in req_fields_r])
        for ind, run_id in enumerate(run_list):
            run = self.get_run_widgets(run_id)
            run.set_flags(req_fields_r[ind])
            self.flag_tab_bar(self.uiRunTabs, f'Run_{run_id}', len(req_fields_r[ind]))
        num_missing_invalid_s = getattr(self, 'num_missing_invalid_s', 0)
        if num_missing_invalid_s == 0:
            self.flag_tab_bar(self.uiMainTabs, 'Sta', num_missing_invalid_r)
        return num_missing_invalid_r

    def validate_run(self, run_id=None):
        """Validate run metadata entry."""
        run_list = self.run_list if run_id is None else [run_id]
        md_fields_r = {f'Run_{run_id}': attrgetter(f'run_{run_id}_widgets.model')(self)
                       for run_id in run_list}
        self.phoenix_md.populate_run_props(md_fields_r, run_list)
        num_missing_invalid_r = self.flag_run_md(run_list)
        return num_missing_invalid_r

    def flag_num_chas_run(self, run_id, cat, req_fields):
        """
        Flag num_chas_run at the elec/mag level if the number of electrode
        pairs/coils does not match the number of electric/magnetic
        channels in components_recorded for a given run
        """
        widget_style = ""
        c_field = f'run_{run_id}_num_{cat.lower()}'
        if c_field in [y for k, v in req_fields.items() for y in v]:
            widget_id = f'QComboBox#{c_field}'
            widget_style = f'{widget_id} {{padding: 10px; border:2px ridge red;}}'
        uiRun = getattr(self, f'uiRun_{run_id}_{cat}')
        uiRun.setStyleSheet(widget_style)

    def flag_other_cha_md(self, run_id, cha_type, num_chas_run, req_fields):
        # Flag every other required electric/magnetic metadata fields that
        # don't pass validation
        num = 0
        for cha_num, req_field in req_fields.items():
            cha = getattr(self, f'run_{run_id}_{cha_type}_{cha_num}_widgets')
            # The electrode pair/coil number for a run ranges from one to
            # the number of electrode pairs/coils deployed during that given
            # run.
            if cha_num in range(1, num_chas_run + 1):
                cha.set_flags(req_field)
                num += len(req_field)
            else:
                cha.set_flags([])
        return num

    def flag_all_cha_md(self, run_list, num_chas, cha_type):
        """
        Flag electric/magnetic metadata fields that are required but unfilled
        or invalid.
        """
        cat = cha_type.capitalize()
        num_missing_invalid_c = 0
        for run_id in run_list:
            num_chas_run = num_chas.get(run_id)
            if num_chas_run is None:
                continue
            req_fields = {x.cha_num: x.md_missing | x.md_invalid
                          for x in self.phoenix_md.filter_cha(cha_type, run_id)}
            self.flag_num_chas_run(run_id, cat, req_fields)
            num = self.flag_other_cha_md(run_id, cha_type, num_chas_run, req_fields)
            num_missing_invalid_c += num
            cha_run_tab = getattr(self, f'uiRunTabs_{cat}')
            self.flag_tab_bar(cha_run_tab, f'Run_{run_id}_{cat}', num)
        self.flag_tab_bar(self.uiMainTabs, cat, num_missing_invalid_c)
        return num_missing_invalid_c

    def validate_cha(self, cha_type, run_id=None):
        """Validate electric/magnetic metadata entry."""
        run_list = self.run_list if run_id is None else [run_id]
        num_chas = {x: getattr(self, f'run_{x}_num_{cha_type}', None) for x in run_list}
        # The variable ind_cha defines the electrode pair or coil number for a
        # given run.
        # The maximum number of electrode pairs/coils are being defined by
        # NUM_CHA_MAX.
        md_fields_c = {f'Run_{run_id}_{cha_type}_{ind_cha}':
                       attrgetter(f'run_{run_id}_{cha_type}_{ind_cha}_widgets.model')(self)
                       for ind_cha in range(1, NUM_CHA_MAX[cha_type] + 1)
                       for run_id in run_list}
        num_chas = self.phoenix_md.populate_chas_props(cha_type, md_fields_c, run_list, num_chas)
        num_missing_invalid_c = self.flag_all_cha_md(run_list, num_chas, cha_type)
        return num_missing_invalid_c

    def set_validation(self):
        """
        Set up validation of user-provided metadata at all levels (net, sta,
        run, elec, mag) and keep track of number of invalid or missing metadata
        field.
        """
        logger.info(f"{'-'*20} Start - Validation of metadata entry set {'-'*20}")
        num_missing_invalid_n = self.validate_net()
        num_missing_invalid_s = self.validate_sta()
        num_missing_invalid_r = self.validate_run()
        num_missing_invalid_e = self.validate_cha('elec')
        num_missing_invalid_m = self.validate_cha('mag')
        logger.info(f"{'-'*20} End - Validation of metadata entry set {'-'*20}")
        self.num_missing_invalid = sum([num_missing_invalid_n, num_missing_invalid_s,
                                        num_missing_invalid_r, num_missing_invalid_e,
                                        num_missing_invalid_m])

    def display_conversion_msg(self):
        """
        Inform users that they can now convert their data and metadata to an
        archivable format by displaying a message in a pop-up window.
        """
        text = "Data and Metadata Conversion Status..."
        details = ("All required metadata are properly filled out.\n"
                   "You can now convert your data and metadata to an "
                   "archivable format!\nConsider saving the inputted "
                   "metadata (⌘S) before proceeding.")
        icon = QtWidgets.QMessageBox.Information
        msg_box = self.create_message_box(text, details, icon)
        ret = msg_box.exec()
        if ret == QtWidgets.QMessageBox.Ok:
            logger.info('All required metadata have been properly filled out '
                        'by the user. Data and metadata conversion actions '
                        'have been enabled.')

    def enable_conversion(self):
        """
        Enable data and metadata conversion icons if all required metadata are
        valid.
        """
        num_missing_invalid = getattr(self, 'num_missing_invalid', None)
        if num_missing_invalid == 0:
            self.actionWrite_mseed.setEnabled(True)
            self.actionWrite_StationXML.setEnabled(True)
            self.display_conversion_msg()
        else:
            self.actionWrite_mseed.setEnabled(False)
            self.actionWrite_StationXML.setEnabled(False)

    def display_mseed_msg(self):
        """
        Inform users that the day-long mseed files were successfully written.
        """
        text = "Data Conversion Complete!"
        details = ("Writing of the daylong mseed files for each channel and "
                   "location code was successfully completed.")
        icon = QtWidgets.QMessageBox.Information
        msg_box = self.create_message_box(text, details, icon)
        msg_box.exec()

    def write_mseed_steps(self):
        """Steps to write day-long mseed files."""
        self.phoenix_data.update_data(md=self.phoenix_md)
        write_daylong_mseed(self.phoenix_data)
        self.display_mseed_msg()

    def display_stationxml_msg(self):
        """
        Inform users that the StationXML files were successfully written.
        """
        text = "Metadata Conversion Complete!"
        details = "Writing of the StationXML file was successfully completed."
        icon = QtWidgets.QMessageBox.Information
        msg_box = self.create_message_box(text, details, icon)
        msg_box.exec()

    def write_stationxml_steps(self):
        """Steps to write StationXML files."""
        self.phoenix_data.update_data(md=self.phoenix_md)
        write_stationxml(self.phoenix_md)
        self.display_stationxml_msg()

    def processing_conversion_steps(self):
        """
        Set up processing and conversion steps. The processing steps consists
        in submitting the metadata entries and validating the user-provided
        metadata at all levels (net, sta, run, elec, magnetic).
        The conversion steps include writing day-long mseed and StationXML files.
        """
        # --- Submit metadata entries ---
        self.uiValidateMD.clicked.connect(self.net_widgets.mapper.submit)
        self.uiValidateMD.clicked.connect(self.sta_widgets.mapper.submit)
        for run_id in self.run_list:
            # -- Run --
            run = self.get_run_widgets(run_id)
            self.uiValidateMD.clicked.connect(run.mapper.submit)
            # -- Electric & Magnetic --
            for cha_type in ['elec', 'mag']:
                for ind in range(1, NUM_CHA_MAX[cha_type] + 1):
                    cha = self.get_cha_widgets(run_id, ind, cha_type)
                    self.uiValidateMD.clicked.connect(cha.mapper.submit)
        # --- Validate metadata entries ---
        self.uiValidateMD.clicked.connect(lambda checked: self.set_validation())
        # --- Convert data and metadata to archivable format ---
        self.uiValidateMD.clicked.connect(lambda checked: self.enable_conversion())
        self.actionWrite_mseed.triggered.connect(lambda checked: self.write_mseed_steps())
        self.actionWrite_StationXML.triggered.connect(lambda checked: self.write_stationxml_steps())

    def close_all_windows(self):
        # Close "Help" window
        self.help_window.close()
        # Close "Electrode - Specs" and "Fluxgate - Specs" windows
        for run_id in self.run_list:
            for cha_type in ['elec', 'mag']:
                for ind in range(1, NUM_CHA_MAX[cha_type] + 1):
                    cha = self.get_cha_widgets(run_id, ind, cha_type)
                    cha.inst_widgets.close()

    def closeEvent(self, event):
        """
        Reimplement closeEvent() event handler to detect when the user attempts
        to close the main window and ask whether they want to save any changes
        made.
        """
        text = "Exiting phoenix2seed application..."
        details = "Do you want to save any pending changes?"
        icon = QtWidgets.QMessageBox.Question
        msg_box = self.create_message_box(text, details, icon)
        msg_box.setStandardButtons(QtWidgets.QMessageBox.Save |
                                   QtWidgets.QMessageBox.Discard |
                                   QtWidgets.QMessageBox.Cancel)
        msg_box.setDefaultButton(QtWidgets.QMessageBox.Save)
        msg_box.setEscapeButton(QtWidgets.QMessageBox.Cancel)
        msg_box.buttonClicked.connect(self.close_all_windows)
        ret = msg_box.exec()
        if ret == QtWidgets.QMessageBox.Save:
            self.save_md()
            event.accept()
        elif ret == QtWidgets.QMessageBox.Discard:
            event.accept()
        else:
            event.ignore()
