# -*- coding: utf-8 -*-
"""
The metadata for a complete  MT dataset is structured into the following
categories: Net, Sta, Run, Elec, and Mag.

This module defines these categories as dataclasses with specific list of
attributes and validation methods.

Maeva Pourpoint - IRIS/PASSCAL
"""

from __future__ import annotations

import re

from dataclasses import dataclass, field
from obspy import UTCDateTime
from typing import Dict, Optional, Set, Union

from phoenix2seed.logging import parse_config_ini, setup_logger
from phoenix2seed.utils import (check_inst_specs, check_sn, ELEC_DEFAULT,
                                COIL_DEFAULT)

# Read config.ini file and set up logging
config = parse_config_ini()
logger = setup_logger(__name__)

COMP_NAMING_CONV = {'elec': {'North-South': 'Ex', 'South-North': 'Ex',
                             'East-West': 'Ey', 'West-East': 'Ey'},
                    'mag': {'North-South': 'Hx', 'East-West': 'Hy',
                            'Vertical': 'Hz'}}
MIN_MAX_RANGE: Dict = {'azimuth': [0, 360], 'dc voltage': [0, 0.1],
                       'declination': [-90, 90], 'dipole length': [0, 99999],
                       'elevation': [-500, 8500], 'latitude': [-90, 90],
                       'longitude': [-180, 180], 'contact resistance': [0, 3000],
                       'tilt': [-90, 90]}
# Placeholder values for the validate_geographics method
MIN_MAX_VALUE: Dict = {'azimuth': [-99999, 99999], 'declination': [-99999, 99999],
                       'tilt': [-99999, 99999]}
MSG_MAP = {'Net': 'network', 'Sta': 'station', 'Run': 'run', 'Elec': 'electric',
           'Mag': 'magnetic'}
SAMPLE_RATE = '24000'  # Native sampling rate in Hz


@dataclass
class BaseMD:
    """
    This class handles the metadata attributes shared at all levels: network;
    station; run; electric; magnetic; auxiliary.
    """
    # List of metadata fields that are invalid
    md_invalid: Set[str] = field(default_factory=set)
    # List of metadata fields that are required for archiving but "missing"
    # because not provided by the user.
    md_missing: Set[str] = field(default_factory=set)


@dataclass
class BaseNet(BaseMD):
    """
    This class handles the metadata attributes shared at the network and station
    levels.
    """
    end: Optional[UTCDateTime] = field(default=None,
                                       metadata={'xml_id': ['end_date'],
                                                 'req': True, 'gui': True})
    start: Optional[UTCDateTime] = field(default=None,
                                         metadata={'xml_id': ['start_date'],
                                                   'req': True, 'gui': True})

    def validate_end(self, md_input: UTCDateTime, data_input: UTCDateTime) -> bool:
        if isinstance(md_input, UTCDateTime) and md_input >= data_input:
            return True
        else:
            logger.error("The end date of the {} should be in UTC and greater "
                         "than the data acquisition end time."
                         .format(MSG_MAP[self.__class__.__name__.split('_')[0]]))
            return False

    def validate_start(self, md_input: UTCDateTime, data_input: UTCDateTime) -> bool:
        if isinstance(md_input, UTCDateTime) and md_input <= data_input:
            return True
        else:
            logger.error("The start date of the {} should be in UTC and lower "
                         "than the data acquisition start time."
                         .format(MSG_MAP[self.__class__.__name__.split('_')[0]]))
            return False


@dataclass
class Net(BaseNet):
    """This class handles all the metadata attributes at the network level."""
    acq_by_author: Optional[str] = field(default=None,
                                         metadata={'xml_id': ['comments', 'mt.network.acquired_by.author'],
                                                   'req': False, 'gui': True})
    acq_by_comments: Optional[str] = field(default=None,
                                           metadata={'xml_id': ['comments', 'mt.network.acquired_by.comments'],
                                                     'req': False, 'gui': True})
    archive_net: Optional[str] = field(default=None,
                                       metadata={'req': True, 'gui': True})
    comments: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.network.comments'],
                                              'req': False, 'gui': True})
    country: Optional[str] = field(default=None,
                                   metadata={'xml_id': ['comments', 'mt.network.country'],
                                             'req': False, 'gui': True})
    geo_name: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.network.geographic_name'],
                                              'req': True, 'gui': True})
    name: Optional[str] = field(default=None,
                                metadata={'xml_id': ['description'],
                                          'req': True, 'gui': True})
    project: Optional[str] = field(default=None,
                                   metadata={'xml_id': ['comments', 'mt.network.project'],
                                             'req': True, 'gui': True})
    project_lead_author: Optional[str] = field(default=None,
                                               metadata={'xml_id': ['operators.contacts.names'],
                                                         'req': False, 'gui': True})
    project_lead_organization: Optional[str] = field(default=None,
                                                     metadata={'xml_id': ['operators.agency'],
                                                               'req': False, 'gui': True})
    summary: Optional[str] = field(default=None,
                                   metadata={'xml_id': ['comments', 'mt.network.summary'],
                                             'req': False, 'gui': True})

    @staticmethod
    def validate_archive_net(md_input: str) -> bool:
        try:
            valid = re.match(r"^\w{2}$", md_input)
        except TypeError:
            logger.error("The network code should be a string.")
            return False
        else:
            if valid is None:
                logger.error("The network code should be two alphanumeric "
                             "character long.")
            return bool(valid)


@dataclass
class BaseSta(BaseMD):
    """
    This class handles the metadata attributes shared at the station and
    channel levels.
    """
    data_type: str = field(default='BBMT',
                           metadata={'xml_id': ['comments', 'mt.station.data_type'],
                                     'req': True, 'gui': False})
    elev: Optional[float] = field(default=None,
                                  metadata={'req': True, 'gui': True})
    lat: Optional[float] = field(default=None,
                                 metadata={'req': True, 'gui': True})
    lon: Optional[float] = field(default=None,
                                 metadata={'req': True, 'gui': True})

    @staticmethod
    def validate_geos(param: tuple, md_input: float) -> bool:
        geo, min_range, max_range, min_val, max_val, val, units = param
        try:
            md_input = float(md_input)
        except (TypeError, ValueError):
            logger.error("{} should be a float.".format(geo.capitalize()))
            return False
        else:
            if md_input < min_range or md_input > max_range:
                logger.error("Unexpected {0}! The {0} should be between {1}"
                             "{3} and {2}{3}.".format(geo, min_range,
                                                      max_range, units))
                return False
            elif md_input < min_val or md_input > max_val:
                tolerance = max_val - val
                logger.warning("Unexpected {0}! Provided {0} should "
                               "roughly match the {0} recorded by the on-site "
                               "GPS ({1:.3f} ± {2:.3f}{3})."
                               .format(geo, val, tolerance, units))
            return True

    def validate_elev(self, md_input: float, data_input: float) -> bool:
        min_val, max_val = [data_input + x for x in [-100, 100]]
        min_range, max_range = MIN_MAX_RANGE['elevation']
        param = ("station elevation", min_range, max_range, min_val, max_val, data_input, 'm')
        return BaseSta.validate_geos(param, md_input)

    def validate_lat(self, md_input: float, data_input: float) -> bool:
        min_val, max_val = [data_input + x for x in [-0.005, 0.005]]
        min_range, max_range = MIN_MAX_RANGE['latitude']
        param = ("station latitude", min_range, max_range, min_val, max_val, data_input, '°')
        return BaseSta.validate_geos(param, md_input)

    def validate_lon(self, md_input: float, data_input: float) -> bool:
        min_val, max_val = [data_input + x for x in [-0.005, 0.005]]
        min_range, max_range = MIN_MAX_RANGE['longitude']
        param = ("station longitude", min_range, max_range, min_val, max_val, data_input, '°')
        return BaseSta.validate_geos(param, md_input)


@dataclass
class Sta(BaseNet, BaseSta):
    """This class handles all the metadata attributes at the station level."""
    archive_id: Optional[str] = field(default=None,
                                      metadata={'req': True, 'gui': True})
    comments: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.station.comments'],
                                              'req': False, 'gui': True})
    declination_model: Optional[str] = field(default=None,
                                             metadata={'xml_id': ['comments', 'mt.station.location.declination.model'],
                                                       'req': True, 'gui': True})
    declination_val: Optional[float] = field(default=None,
                                             metadata={'xml_id': ['comments', 'mt.station.location.declination.value'],
                                                       'req': True, 'gui': True})
    geo_name: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['site'],
                                              'req': True, 'gui': True})
    orientation_method: Optional[str] = field(default=None,
                                              metadata={'xml_id': ['comments', 'mt.station.orientation.method'],
                                                        'req': False, 'gui': True})
    orientation_reference_frame: Optional[str] = field(default=None,
                                                       metadata={'xml_id': ['comments', 'mt.station.orientation.reference_frame'],  # noqa: E501
                                                                 'req': True, 'gui': True})
    run_list: str = field(default='a',
                          metadata={'xml_id': ['comments', 'mt.station.run_list'],
                                    'req': True, 'gui': False})
    software_author: str = field(default=config['software']['author'],
                                 metadata={'xml_id': ['comments', 'mt.station.provenance.software.author'],
                                           'req': False, 'gui': False})
    software_name: str = field(default=config['software']['name'],
                               metadata={'xml_id': ['comments', 'mt.station.provenance.software.name'],
                                         'req': False, 'gui': False})
    software_version: str = field(default=config['software']['version'],
                                  metadata={'xml_id': ['comments', 'mt.station.provenance.software.version'],
                                            'req': False, 'gui': False})
    submitter_author: Optional[str] = field(default=None,
                                            metadata={'xml_id': ['comments', 'mt.station.provenance.submitter.author'],
                                                      'req': False, 'gui': True})
    submitter_organization: Optional[str] = field(default=None,
                                                  metadata={'xml_id': ['comments', 'mt.station.provenance.submitter.organization'],  # noqa: E501
                                                            'req': False, 'gui': True})

    @staticmethod
    def validate_archive_id(md_input: str) -> bool:
        try:
            valid = re.match(r"^\w{3,5}$", md_input)
        except TypeError:
            logger.error("The station name should be a string.")
            return False
        else:
            if valid is None:
                logger.error("The station name should be between three and "
                             "five alphanumeric character long.")
            return bool(valid)

    def validate_declination_val(self, md_input: float) -> bool:
        min_val, max_val = MIN_MAX_VALUE['declination']
        min_range, max_range = MIN_MAX_RANGE['declination']
        param = ("declination angle", min_range, max_range, min_val, max_val, None, '°')
        return BaseSta.validate_geos(param, md_input)


@dataclass
class Run(BaseMD):
    """This class handles all the metadata attributes at the run level."""
    acq_by_author: Optional[str] = field(default=None,
                                         metadata={'xml_id': ['comments', 'mt.run:a.acquired_by.author'],
                                                   'req': False, 'gui': True})
    acq_by_comments: Optional[str] = field(default=None,
                                           metadata={'xml_id': ['comments', 'mt.run:a.acquired_by.comments'],
                                                     'req': False, 'gui': True})
    chas_rec: Optional[str] = field(default=None,
                                    metadata={'req': True, 'gui': True})
    comments: Optional[str] = field(default=None,
                                    metadata={'xml_id': ['comments', 'mt.run:a.comments'],
                                              'req': False, 'gui': True})
    data_type: str = field(default='BBMT',
                           metadata={'xml_id': ['type'],
                                     'req': True, 'gui': False})
    datalogger_manufacturer: str = field(default='Phoenix Geophysics',
                                         metadata={'xml_id': ['manufacturer'],
                                                   'req': True, 'gui': False})
    datalogger_model: str = field(default='MTU-5C',
                                  metadata={'xml_id': ['model'],
                                            'req': True, 'gui': True})
    datalogger_sn: Optional[str] = field(default=None,
                                         metadata={'xml_id': ['serial_number'],
                                                   'req': True, 'gui': True})
    datalogger_type: str = field(default='type: ultra wideband',
                                 metadata={'xml_id': ['description'],
                                           'req': False, 'gui': False})
    end: Optional[UTCDateTime] = field(default=None,
                                       metadata={'xml_id': ['removal_date'],
                                                 'req': True, 'gui': False})
    md_by_author: Optional[str] = field(default=None,
                                        metadata={'xml_id': ['comments', 'mt.run:a.metadata_by.author'],
                                                  'req': False, 'gui': True})
    md_by_comments: Optional[str] = field(default=None,
                                          metadata={'xml_id': ['comments', 'mt.run:a.metadata_by.comments'],
                                                    'req': False, 'gui': True})
    resource_id: Optional[str] = field(default=None,
                                       metadata={'xml_id': ['resource_id'],
                                                 'req': True, 'gui': False})
    sample_rates: str = field(default=SAMPLE_RATE,
                              metadata={'req': True, 'gui': True})
    start: Optional[UTCDateTime] = field(default=None,
                                         metadata={'xml_id': ['installation_date'],
                                                   'req': True, 'gui': False})

    @property
    def run_id(self) -> str:
        return self.resource_id.split(':')[1]  # type: ignore

    def validate_datalogger_sn(self, md_input: Union[str, int], data_input: str) -> bool:
        equipment = "data logger (run '{}')".format(self.run_id)
        return check_sn(equipment, str(md_input), data_input)


@dataclass
class BaseChannel(BaseSta):
    """
    This class handles the metadata attributes shared at the electric and
    magnetic levels.
    """
    cha_port: Optional[str] = field(default=None,
                                    metadata={'req': True, 'gui': True})
    comp: Optional[str] = field(default=None,
                                metadata={'xml_id': ['alternate_code'],
                                          'req': True, 'gui': False})
    data_type: str = field(default='GEOPHYSICAL',
                           metadata={'xml_id': ['types'],
                                     'req': True, 'gui': False})
    depth: int = field(default=0,
                       metadata={'req': True, 'gui': False})
    inst_manufacturer: Optional[str] = field(default=None,
                                             metadata={'req': True, 'gui': True})
    inst_model: Optional[str] = field(default=None,
                                      metadata={'req': True, 'gui': True})
    inst_specs: Optional[str] = field(default=None,
                                      metadata={'req': True, 'gui': True})
    inst_type: Optional[str] = field(default=None,
                                     metadata={'req': False, 'gui': True})
    loc_code: str = field(default='',
                          metadata={'req': True, 'gui': False})
    meas_azimuth: Optional[float] = field(default=None,
                                          metadata={'xml_id': ['azimuth'],
                                                    'req': True, 'gui': True})
    meas_tilt: Optional[float] = field(default=None,
                                       metadata={'xml_id': ['dip'],
                                                 'req': True, 'gui': True})
    cha_num: Optional[int] = field(default=None,
                                   metadata={'req': True, 'gui': False})
    run_id: Optional[str] = field(default=None,
                                  metadata={'req': True, 'gui': False})

    def validate_inst_specs(self, md_input: str) -> bool:
        if self.__class__.__name__.split('_')[0] == 'Elec':
            equipment = 'electrode'
            cha_info = "for electrode pair {} (run '{}')".format(self.cha_num, self.run_id)
        else:
            equipment = 'coil'
            cha_info = "for coil {} (run '{}')".format(self.cha_num, self.run_id)
        return check_inst_specs(md_input, equipment, cha_info)

    def validate_meas_azimuth(self, md_input: float) -> bool:
        if isinstance(self, Elec):
            geo = "azimuth angle of electrode pair {} (run '{}')".format(self.cha_num, self.run_id)
        if isinstance(self, Mag):
            geo = "azimuth angle of coil {} (run '{}')".format(self.cha_num, self.run_id)
        min_val, max_val = MIN_MAX_VALUE['azimuth']
        min_range, max_range = MIN_MAX_RANGE['azimuth']
        param = (geo, min_range, max_range, min_val, max_val, None, '°')
        return BaseSta.validate_geos(param, md_input)

    def validate_meas_tilt(self, md_input: float) -> bool:
        if isinstance(self, Elec):
            geo = "tilt angle of electrode pair {} (run '{}')".format(self.cha_num, self.run_id)
        if isinstance(self, Mag):
            geo = "tilt angle of coil {} (run '{}')".format(self.cha_num, self.run_id)
        min_val, max_val = MIN_MAX_VALUE['tilt']
        min_range, max_range = MIN_MAX_RANGE['tilt']
        param = (geo, min_range, max_range, min_val, max_val, None, '°')
        return BaseSta.validate_geos(param, md_input)


@dataclass
class Elec(BaseChannel):
    """This class handles all the metadata attributes at the electric level."""
    contact_resistance_end: Optional[float] = field(default=None,
                                                    metadata={'req': False, 'gui': True})
    contact_resistance_start: Optional[float] = field(default=None,
                                                      metadata={'req': False, 'gui': True})
    dc_end: Optional[float] = field(default=None,
                                    metadata={'req': False, 'gui': True})
    dc_start: Optional[float] = field(default=None,
                                      metadata={'req': False, 'gui': True})
    dipole_len: Optional[float] = field(default=None,
                                        metadata={'req': True, 'gui': True})
    neg_elec_dir: Optional[str] = field(default=None,
                                        metadata={'req': True, 'gui': True})
    neg_elec_sn: Optional[str] = field(default=None,
                                       metadata={'req': True, 'gui': True})
    pos_elec_dir: Optional[str] = field(default=None,
                                        metadata={'req': True, 'gui': True})
    pos_elec_sn: Optional[str] = field(default=None,
                                       metadata={'req': True, 'gui': True})

    def set_comp(self) -> None:
        """
        Set component for a given electric channel based on the direction of
        its electrode pair.
        """
        try:
            dir_ = "-".join([self.pos_elec_dir, self.neg_elec_dir])  # type: ignore
        except TypeError:
            logger.error("The direction of the positive and negative "
                         "electrodes should be either: North, South, East or "
                         "West. Please, provide electrode direction(s) for "
                         "electrode pair {} (run '{}')!"
                         .format(self.cha_num, self.run_id))
            self.md_invalid.update({'pos_elec_dir', 'neg_elec_dir'})
        else:
            comp = COMP_NAMING_CONV['elec'].get(dir_)
            if comp is None:
                if self.pos_elec_dir == self.neg_elec_dir:
                    logger.error("The direction of the positive and negative "
                                 "electrodes in a given pair cannot be the "
                                 "same (check run '{}' and electrode pair {})!"
                                 .format(self.run_id, self.cha_num))
                logger.warning("By convention, one electrode pair is installed "
                               "in a north-south direction and the other pair "
                               "in a east-west direction (check run '{}' and "
                               "electrode pair {})!".format(self.run_id, self.cha_num))
                self.md_invalid.update({'pos_elec_dir', 'neg_elec_dir'})
            else:
                self.comp = comp
                self.md_invalid.discard('pos_elec_dir')
                self.md_invalid.discard('neg_elec_dir')

    def set_info(self) -> None:
        """
        Set electrode manufacturer, model and type based on user provided
        electrode specs.
        """
        if self.inst_specs == ELEC_DEFAULT:
            self.inst_manufacturer = "Borin"
            self.inst_model = "STELTH 4"
            self.inst_type = "Silver-Silver Chloride"
        elif self.inst_specs is not None:
            specs = [x.split(':')[-1].strip() for x in self.inst_specs.split('-')]
            self.inst_manufacturer, self.inst_model, self.inst_type = specs

    def validate_e_prop(self, param: tuple, md_input: float) -> bool:
        e_prop, min_range, max_range, units = param
        try:
            md_input = float(md_input)
        except (TypeError, ValueError):
            logger.error("The {} for electrode pair {} (run '{}') should be "
                         "a float.".format(e_prop, self.cha_num, self.run_id))
            return False
        else:
            if md_input < min_range:
                logger.error("The {} for electrode pair {} (run '{}') should "
                             "be positive.".format(e_prop, self.cha_num, self.run_id))
                return False
            if md_input > max_range:
                msg = ''
                if 'resistance' in e_prop or 'voltage' in e_prop:
                    msg = ("A {} > {}{} is indicative of poor contact."
                           .format(e_prop, max_range, units))
                logger.warning("The {} for electrode pair {} (run '{}') should "
                               "be less than {}{}. {}"
                               .format(e_prop, self.cha_num, self.run_id,
                                       max_range, units, msg))
            return True

    def validate_contact_resistance_end(self, md_input: float) -> bool:
        if md_input is not None:
            min_range, max_range = MIN_MAX_RANGE['contact resistance']
            param = ("contact resistance (end)", min_range, max_range, 'Ω')
            return self.validate_e_prop(param, md_input)
        return True

    def validate_contact_resistance_start(self, md_input: float, data_input: Dict) -> bool:
        if data_input.get(self.cha_port) is not None:
            rec_resistance_p, rec_resistance_n = data_input[self.cha_port].values()
            avg_rec_resistance = (rec_resistance_p + rec_resistance_n)/2
            logger.info("The contact resistance (start) for electrode pair {} "
                        "(run '{}') is rougly {:.3f} Ω as measured by the MTU-5C "
                        "receiver.".format(self.cha_num, self.run_id, avg_rec_resistance))
        if md_input is not None:
            min_range, max_range = MIN_MAX_RANGE['contact resistance']
            param = ("contact resistance (start)", min_range, max_range, 'Ω')
            return self.validate_e_prop(param, md_input)
        else:
            return True

    def validate_dc_end(self, md_input: float) -> bool:
        if md_input is not None:
            min_range, max_range = MIN_MAX_RANGE['dc voltage']
            param = ("DC voltage (end)", min_range, max_range, 'V')
            return self.validate_e_prop(param, md_input)
        return True

    def validate_dc_start(self, md_input: float) -> bool:
        if md_input is not None:
            min_range, max_range = MIN_MAX_RANGE['dc voltage']
            param = ("DC voltage (start)", min_range, max_range, 'V')
            return self.validate_e_prop(param, md_input)
        return True

    def validate_dipole_len(self, md_input: float, data_input: Dict) -> bool:
        min_range, max_range = MIN_MAX_RANGE['dipole length']
        param = ("dipole length", min_range, max_range, 'm')
        if not self.validate_e_prop(param, md_input):
            return False
        rec_dipole_len = data_input[self.cha_port]
        if not (float(md_input) == rec_dipole_len):
            logger.warning("The dipole length for electrode pair {} (run '{}') "
                           "does not match the dipole length value that was "
                           "configured using EMpower ({} != {})!"
                           .format(self.cha_num, self.run_id, md_input, rec_dipole_len))
        return True


@dataclass
class Mag(BaseChannel):
    """This class handles all the metadata attributes at the magnetic level."""
    coil_dir: Optional[str] = field(default=None,
                                    metadata={'req': True, 'gui': True})
    coil_sn: Optional[str] = field(default=None,
                                   metadata={'req': True, 'gui': True})

    def set_comp(self) -> None:
        """
        Set component for a given magnetic channel based on the direction of
        its coil.
        """
        if self.coil_dir is None:
            logger.error("The direction of the coil should either be: "
                         "North-South, East-West or Vertical. Please provide "
                         "the direction of coil {} (run '{}')!"
                         .format(self.cha_num, self.run_id))
            self.md_invalid.update({'coil_dir'})
        else:
            comp = COMP_NAMING_CONV['mag'].get(self.coil_dir)
            self.comp = comp
            self.md_invalid.discard('coil_dir')

    def set_info(self) -> None:
        if self.inst_specs in COIL_DEFAULT.keys():
            self.inst_manufacturer = COIL_DEFAULT[self.inst_specs]['manufacturer']  # type: ignore
            self.inst_model = COIL_DEFAULT[self.inst_specs]['model']  # type: ignore
            self.inst_type = COIL_DEFAULT[self.inst_specs]['type']  # type: ignore
        elif self.inst_specs is not None:
            specs = [x.split(':')[-1].strip() for x in self.inst_specs.split('-')]
            self.inst_manufacturer, self.inst_model, self.inst_type = specs

    def validate_coil_sn(self, md_input: Union[str, int], data_input: Dict) -> bool:
        equipment = "coil {} (run '{}')".format(self.cha_num, self.run_id)
        return check_sn(equipment, str(md_input), data_input.get(self.cha_port))
