#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Routines to convert the provided metadata to StationXML files.

Maeva Pourpoint - EPIC
"""

from __future__ import annotations

import sys

from dataclasses import fields
from obspy import UTCDateTime
from obspy.core.inventory import (Channel, Comment, Equipment, Inventory,
                                  Network, Operator, Person, Site, Station)
from typing import Dict, List, Optional, Set, Tuple, TYPE_CHECKING

from phoenix2seed.phoenix_data import CHA_NAMING_CONV
from phoenix2seed.phoenix_response import PhoenixResponse, CHA_UNITS
from phoenix2seed.logging import parse_config_ini, setup_logger
from phoenix2seed.utils import str2list, is_empty

if TYPE_CHECKING:
    from dataclasses import Field
    from phoenix2seed.phoenix_metadata import (PhoenixMetadata, Net, Sta, Run, Elec,
                                               Mag, DCS, DCS_SUB_EM, MD_VAL)
    from obspy.core.inventory import Response

# Read config.ini file and set up logging
config = parse_config_ini()
logger = setup_logger(__name__)


def create_inventory() -> Inventory:
    """Create Inventory object."""
    software_module = f"{config['software']['name']}.{config['software']['version']}"
    inv = Inventory(module=software_module,
                    module_uri='www.passcal.nmt.edu',
                    source='EPIC')
    return inv


def get_xml_field(cat: DCS, field: Field) -> Tuple:
    """
    Get xml_id from MappingProxyType metadata for given dataclass attribute.
    Get metadata field value for given dataclass attribute.
    """
    field_info = field.metadata.get('xml_id')
    tmp = getattr(cat, field.name)
    field_val = tmp if not is_empty(tmp) else ''
    return field_info, field_val


def get_comment(field_info: List[str], field_val: MD_VAL, *, ind_run: Optional[int] = None) -> Comment:
    """Create Comment object for a given comment entry."""
    comment = Comment('')
    comment.value = str(field_val)
    subject = field_info[1]
    if ind_run is not None:
        run_id = chr(ord('a') + ind_run)
        subject = subject.replace(':a', f':{run_id}')
    comment.subject = subject
    return comment


def create_net(net: Net) -> Network:
    """Create Network object and populate with network related metadata."""
    comments = []
    person = Person()
    net_ = Network(code=net.archive_net.upper())  # type: ignore
    net_fields = [x for x in fields(net) if x.metadata.get('xml_id') is not None]
    for net_field in net_fields:
        net_field_info, net_field_val = get_xml_field(net, net_field)
        if net_field_info[0] == 'comments' and net_field_val != '':
            comment = get_comment(net_field_info, net_field_val)
            comments.append(comment)
        elif 'operators' in net_field_info[0]:
            if 'agency' in net_field_info[0]:
                # Address agency related change in StationXML 1.1
                if not net_field_val:
                    net_field_val = ' '
                operator = Operator(agency=net_field_val)
            else:
                person.names = str2list(net_field_val)
        elif net_field_info[0] == 'identifiers':
            net_.identifiers = str2list(net_field_val)
        else:
            setattr(net_, net_field_info[0], net_field_val)
    operator.contacts = [person]
    net_.operators = [operator]
    net_.comments = comments
    return net_


def create_sta(sta: Sta, runs: List[Run]) -> Station:
    """Create Station object and populate with station related metadata."""
    comments = []
    sta_ = Station(code=sta.archive_id.upper(), latitude=sta.lat,  # type: ignore
                   longitude=sta.lon, elevation=sta.elev)
    sta_fields = [x for x in fields(sta) if x.metadata.get('xml_id') is not None]
    # --- Metadata fields at the station level ---
    for sta_field in sta_fields:
        sta_field_info, sta_field_val = get_xml_field(sta, sta_field)
        if sta_field_info[0] == 'comments' and sta_field_val != '':
            comment = get_comment(sta_field_info, sta_field_val)
            comments.append(comment)
        elif sta_field_info[0] == 'site':
            # Temporary response flag to identify station with incorrect archived response
            resp_flag = ' ("dummy" instrument response)'
            sta_.site = Site(sta_field_val + resp_flag)
        else:
            setattr(sta_, sta_field_info[0], sta_field_val)
    # --- Comments at the run level ---
    for ind_run, run in enumerate(runs):
        run_fields = [x for x in fields(run)
                      if x.metadata.get('xml_id') is not None
                      and 'comments' in x.metadata.get('xml_id')]  # type: ignore
        for run_field in run_fields:
            run_field_info, run_field_val = get_xml_field(run, run_field)
            if run_field_val != '':
                comment = get_comment(run_field_info, run_field_val, ind_run=ind_run)
                comments.append(comment)
    sta_.comments = comments
    return sta_


def create_logger(runs: List[Run]) -> List[Equipment]:
    """Create Equipment object and populate with run related metadata."""
    loggers = []
    for run in runs:
        run_fields = [x for x in fields(run)
                      if x.metadata.get('xml_id') is not None
                      and 'comments' not in x.metadata.get('xml_id')]  # type: ignore
        logger = Equipment()
        for run_field in run_fields:
            run_field_info, run_field_val = get_xml_field(run, run_field)
            setattr(logger, run_field_info[0], run_field_val)
        loggers.append(logger)
    return loggers


def get_id_changes(runs: List) -> Set[str]:
    """
    Get run id if a change at given channel port was made.
    """
    keys = set()
    for x, y in runs:
        keys.update(list(y.keys()))
    changes = set()
    for x, y in zip(runs[:-1], runs[1:]):
        for key in keys:
            port1 = x[1].get(key)
            port2 = y[1].get(key)
            if (port1 and port2) and (port1 != port2):
                changes.add(y[0])
    return changes


def detect_changes_coil(phoenix_md: PhoenixMetadata) -> Set[str]:
    """
    At a given channel port, detect changes in the serial number
    of the coil connected to that port.
    """
    runs_coil_sn = []
    for run_id in phoenix_md.run_list:
        num_m_pairs = len(phoenix_md.get_chas_rec('H', run_id))
        chas = phoenix_md.filter_cha('mag', run_id)[:num_m_pairs]
        runs_coil_sn.append((run_id, {x.cha_port: x.coil_sn for x in chas}))  # type: ignore
    changes = get_id_changes(runs_coil_sn)
    return changes


def detect_changes_datalogger(phoenix_md: PhoenixMetadata) -> Set[str]:
    """Detect changes in data logger serial number."""
    runs_logger_sn = [(x.run_id, x.datalogger_sn) for x in phoenix_md.run]
    changes = {y[0] for x, y in zip(runs_logger_sn[:-1], runs_logger_sn[1:]) if x[1] != y[1]}
    return changes


def detect_changes_dipole_length(phoenix_md: PhoenixMetadata) -> Set[str]:
    """
    At a given channel port, detect changes in the dipole length
    of the electrode pair connected to that port.
    """
    runs_dipole_length = []
    for run_id in phoenix_md.run_list:
        num_e_pairs = len(phoenix_md.get_chas_rec('E', run_id))
        chas = phoenix_md.filter_cha('elec', run_id)[:num_e_pairs]
        runs_dipole_length.append((run_id, {x.cha_port: x.dipole_len for x in chas}))  # type: ignore
    changes = get_id_changes(runs_dipole_length)
    return changes


def detect_epochs(phoenix_md: PhoenixMetadata) -> List:
    """
    Detect epoch(s) based on changes in:
    - coil serial number
    - data logger serial number
    - dipole length of electrode pair
    """
    elec_changes = detect_changes_dipole_length(phoenix_md)
    mag_changes = detect_changes_coil(phoenix_md)
    datalogger_changes = detect_changes_datalogger(phoenix_md)
    changes = datalogger_changes | elec_changes | mag_changes
    changes.add(phoenix_md.run_list[0])
    changes.add(chr(ord(phoenix_md.run_list[-1]) + 1))
    changes = sorted(changes)  # type: ignore
    epochs = []
    e_ports = {k: list(v.keys()) for k, v in phoenix_md.get_cha_infos('elec').items()}
    m_ports = {k: list(v.keys()) for k, v in phoenix_md.get_cha_infos('mag').items()}
    run_ports = {k: e_ports.get(k, []) + m_ports.get(k, []) for k in e_ports.keys() | m_ports.keys()}
    for run_id_start, run_id_end in zip(changes[:-1], changes[1:]):  # type: ignore
        run_ids = [chr(x) for x in range(ord(run_id_start), ord(run_id_end))]
        run_info = {x: run_ports[x] for x in run_ids}
        run_time_period = {run.run_id: (run.start, run.end) for run in phoenix_md.run if run.run_id in run_ids}
        epochs.append([run_info, run_time_period])
    return epochs


def get_e_cha_updates(chas: List[Elec], epoch: Dict) -> Dict:
    """
    Get metadata fields specific to the electric channels (DC voltage,
    contact resistance and dipole length) for each epoch.
    """
    updates: Dict = {x.cha_port: {} for x in chas}
    for x in chas:
        if x.cha_port in epoch[x.run_id]:
            updates[x.cha_port].update({x.run_id: [x.dc_start,
                                                   x.dc_end,
                                                   x.contact_resistance_start,
                                                   x.contact_resistance_end,
                                                   x.dipole_len]})
    return updates


def create_elec_sensor(elec: Elec) -> Equipment:
    """Create Sensor object and populate with electric related metadata."""
    sensor = Equipment()
    sensor.type = 'electrode'
    sensor.manufacturer = elec.inst_manufacturer
    sensor.model = " - ".join([elec.inst_model, elec.inst_type])  # type: ignore
    pos_sn = ": ".join(['positive', str(elec.pos_elec_sn)])
    neg_sn = ": ".join(['negative', str(elec.neg_elec_sn)])
    sensor.serial_number = ", ".join([pos_sn, neg_sn])
    return sensor


def create_mag_sensor(mag: Mag) -> Equipment:
    """Create Sensor object and populate with magnetic related metadata."""
    sensor = Equipment()
    sensor.type = 'coil'
    sensor.description = mag.inst_type
    sensor.manufacturer = mag.inst_manufacturer
    sensor.model = mag.inst_model
    sensor.serial_number = mag.coil_sn
    return sensor


def init_cha(cat: DCS_SUB_EM, sample_rate: int) -> Channel:
    """Instantiate Channel object."""
    cha = Channel(code=CHA_NAMING_CONV[cat.comp][sample_rate],   # type: ignore
                  location_code=cat.loc_code,
                  latitude=cat.lat, longitude=cat.lon,
                  elevation=cat.elev, depth=cat.depth,
                  sample_rate=sample_rate)
    return cha


def create_cha(cha: DCS_SUB_EM, cha_type: str, sample_rate: int) -> Channel:
    """
    Create Channel object and populate class instance using the collected
    metadata fields at the Elec and Mag levels.
    Calls create_elec_sensor or create_mag_sensor based on channel type.
    """
    comments = []
    cha_ = init_cha(cha, sample_rate)
    cha_fields = [x for x in fields(cha) if x.metadata.get('xml_id') is not None]
    for cha_field in cha_fields:
        cha_field_info, cha_field_val = get_xml_field(cha, cha_field)
        if cha_field_info[0] == 'comments' and cha_field_val != '':
            comment = get_comment(cha_field_info, cha_field_val)
            comments.append(comment)
        elif cha_field_info[0] == 'types':
            cha_.types = [cha_field_val]
        else:
            setattr(cha_, cha_field_info[0], cha_field_val)
        cha_.comments = comments
    create_sensor_method = '_'.join(['create', cha_type, 'sensor'])
    sensor = getattr(sys.modules[__name__], create_sensor_method)(cha)
    cha_.sensor = sensor
    return cha_


def update_dates(cha: Channel, dates: List) -> Channel:
    """Update the start and end dates of the channel."""
    cha.start_date = min(dates)[0]
    cha.end_date = max(dates)[1]
    return cha


def update_calibration_units(cha: Channel) -> Channel:
    """Update the calibration units of the channel."""
    cha.calibration_units = CHA_UNITS[0]
    cha.calibration_units_description = CHA_UNITS[1]
    return cha


def add_run_ids_comment(cha: Channel, run_ids: List) -> Channel:
    """Add comment listing the id of the runs for given epoch"""
    comment = Comment('')
    comment.value = ", ".join(run_ids)
    comment.subject = 'mt.run.id'
    cha.comments.append(comment)
    return cha


def add_voltage_resistance_comment(cha: Channel, cha_port: str, e_cha_updates: Dict) -> Channel:
    """
    Add comments about the measured voltages and contact resistances for all
    runs in a given epoch.
    """
    comment_subjects = {0: ['mt.electric.dc.start', 'V'],
                        1: ['mt.electric.dc.end', 'V'],
                        2: ['mt.electric.contact_resistance.start', 'Ω'],
                        3: ['mt.electric.contact_resistance.end', 'Ω']}
    for ind, val in comment_subjects.items():
        subject, unit = val
        val = ' - '.join([f'{k}: {v[ind]}{unit}' for k, v in e_cha_updates[cha_port].items() if v[ind]])  # type: ignore
        if not val:
            continue
        comment = Comment('')
        comment.value = val
        comment.subject = subject
        cha.comments.append(comment)
    return cha


def add_sensor_description(cha: Channel, cha_port: str, e_cha_updates: Dict) -> Channel:
    """"
    Add dipole length information for all runs in a given epoch to sensor
    description.
    """
    val = ' - '.join([f'{k}: {v[-1]} m' for k, v in e_cha_updates[cha_port].items() if v[-1]])
    cha.sensor.description = val
    return cha


def update_cha(cha: Channel, cha_port: str, epoch: List[Dict], e_cha_updates: Dict) -> Channel:
    """
    Update Channel attributes with information for a given epoch:
    start date; end date; calibration units; list of run ids; voltage and contact
    resistance for all runs (only applicable for electric channels).
    """
    run_info, run_time_period = epoch
    run_ids = [key for key, val in run_info.items() if cha_port in val]
    dates = [val for key, val in run_time_period.items() if key in run_ids]
    cha = update_dates(cha, dates)
    cha = update_calibration_units(cha)
    cha = add_run_ids_comment(cha, run_ids)
    if e_cha_updates:
        cha = add_voltage_resistance_comment(cha, cha_port, e_cha_updates)
        cha = add_sensor_description(cha, cha_port, e_cha_updates)
    return cha


def get_cha_resp(sample_rate: int) -> Response:
    """
    Get single stage "dummy" channel response for receiver/sensor system.
    """
    phoenix_resp = PhoenixResponse(sample_rate)
    cha_resp = phoenix_resp.get_cha_resp()
    return cha_resp


def create_all_chas(phoenix_md: PhoenixMetadata, cha_type: str, epochs: List) -> List[Channel]:
    """
    For a given channel type (electric or magnetic), create Channel objects
    for each epoch, set of recorded components and sample rate.
    """
    chas: List = []
    md_chas = getattr(phoenix_md, cha_type)
    for epoch in epochs:
        archived_chas = []
        md_chas_epoch = [x for x in md_chas if x.run_id in epoch[0].keys()]
        e_cha_updates = get_e_cha_updates(md_chas_epoch, epoch[0]) if cha_type == 'elec' else {}
        for md_cha in md_chas_epoch:
            run_id = md_cha.run_id
            if md_cha.cha_port not in epoch[0][run_id]:
                continue
            samples_rates = [int(x) for x in phoenix_md.get_run(run_id).sample_rates.split('-')]
            for sample_rate in samples_rates:
                flag = f'{md_cha.cha_port}_{sample_rate}'
                if flag in archived_chas:
                    continue
                archived_chas.append(flag)
                cha = create_cha(md_cha, cha_type, sample_rate)
                cha = update_cha(cha, md_cha.cha_port, epoch, e_cha_updates)  # type: ignore
                cha.response = get_cha_resp(sample_rate)  # type: ignore
                chas.append(cha)
    return chas


def write_stationxml(phoenix_md: PhoenixMetadata) -> None:
    """
    Create a StationXML file for given station.
    Naming convention for StationXML files: net.sta.project.yyyyjulday.time.xml
    where:
    - net is the network code
    - project is the experiment name
    - yyyy, julday and time is UTC year, julian day and time
    """
    filename = '.'.join([phoenix_md.data_stats['net'], phoenix_md.data_stats['sta'],
                         '_'.join([x.upper() for x in phoenix_md.net.project.split(" ")]),  # type: ignore
                         UTCDateTime().strftime('%Y%j.%H%M%S'), 'xml'])
    filename = str(phoenix_md.output_xml.joinpath(filename))
    logger.info("Writing StationXML file: {}.".format(filename))
    inv = create_inventory()
    net = create_net(phoenix_md.net)
    sta = create_sta(phoenix_md.sta, phoenix_md.run)
    loggers = create_logger(phoenix_md.run)
    epochs = detect_epochs(phoenix_md)
    chas = [c for cha_type in ['elec', 'mag'] for c in create_all_chas(phoenix_md, cha_type, epochs)]
    sta.channels = chas
    sta.equipments = loggers
    net.stations = [sta]
    inv.networks = [net]
    inv.write(filename, format="STATIONXML")
    logger.info("Writing of StationXML file successfully completed!")
