#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Convert Phoenix (MTU-5C) data and metadata into day-long mseed and StationXML
files.

Maeva Pourpoint - IRIS/PASSCAL
"""

import argparse
import re
import sys

from pathlib import Path
from PySide6 import QtWidgets
from typing import Any, Dict, List, Optional, Tuple

from phoenix2seed.data_conversion import write_daylong_mseed
from phoenix2seed.gui_view import PhoenixWindow
from phoenix2seed.logging import initiate_logger, parse_config_ini, setup_logger
from phoenix2seed.phoenix_data import PhoenixData
from phoenix2seed.phoenix_metadata import PhoenixMetadata
from phoenix2seed.utils import is_empty, NUM_CHA_MAX

# Read config.ini file and set up logging
config = parse_config_ini()
logger = setup_logger(__name__)
initiate_logger(logger, config['software']['version'])


def setup_args() -> argparse.ArgumentParser:
    """Set up command line interface."""
    parser = argparse.ArgumentParser(prog='phoenix2seed',
                                     formatter_class=argparse.RawTextHelpFormatter,
                                     description='''
Convert Phoenix data and metadata to day-long mseed and StationXML files.
phoenix2seed can run under three distinct modes:
- quality control mode ('phoenix2seed -d path2data -qc')
- gui only mode ('phoenix2seed -d path2data -g)
- field sheet and gui mode (phoenix2seed -d path2data -m path2metadata)
The 'quality control' mode allows to quickly convert Phoenix data to mseed. The GUI and metadata handling are bypassed.
In the 'gui only' mode, metadata are entered/modified using only the GUI.
In the 'field sheet and gui' mode, metadata are prepopulated from parsing the user-provided field sheets and modified using the GUI.''')  # noqa: E501
    parser.add_argument('-d',
                        dest='path2data',
                        metavar='path2data',
                        help='''Path to the station directory where all folders/files generated by the MTU-5C are stored''')  # noqa: E501
    parser.add_argument('-m',
                        dest='path2md',
                        metavar='path2md',
                        help='''Path to the directory where all Phoenix field spreadsheets are stored''')
    parser.add_argument('-s',
                        dest='path2savedmd',
                        metavar='path2savedmd',
                        help='''Path to the .pkl file where metadata were previously saved''')
    parser.add_argument('-od',
                        dest='output_mseed',
                        metavar='output_mseed',
                        default='MSEED',
                        help='''Output directory name for day-long mseed files''')
    parser.add_argument('-om',
                        dest='output_xml',
                        metavar='output_xml',
                        default='STATIONXML',
                        help='''Output directory name for StationXML files''')
    parser.add_argument('-qc',
                        dest='qc_mode',
                        action='store_true',
                        default=False,
                        help='''Runs phoenix2seed in quality control mode. Handling of metadata is bypassed.''')
    parser.add_argument('-g',
                        dest='gui_only_mode',
                        action='store_true',
                        default=False,
                        help='''Runs phoenix2seed in gui only mode. All metadata are entered using the GUI
(bypassing the use of field sheets).''')
    return parser


def check_path2data(args: argparse.Namespace) -> Path:
    """Check path to data files."""
    if args.path2data and Path(args.path2data).is_dir():
        path2data = Path(args.path2data).resolve()
    else:
        logger.error("Path to Phoenix data not provided! Required input! "
                     "Try --help option.")
        sys.exit(1)
    return path2data


def check_running_mode(args: argparse.Namespace) -> Tuple[bool, bool]:
    """Check running mode selected by end-user."""
    qc_only = args.qc_mode
    gui_only = args.gui_only_mode
    if all([qc_only, gui_only]):
        logger.error("phoenix2seed cannot run under both the 'quality control' "
                     "and 'gui only' mode at the same time! Try --help option "
                     "or choose one or the other mode.")
        sys.exit(1)
    return qc_only, gui_only


def check_path2metadata(args: argparse.Namespace, qc_only: bool, gui_only: bool) -> Optional[Path]:
    """Check path to metadata file(s)."""
    if args.path2md is None:
        if not any([qc_only, gui_only]):
            logger.error("No metadata path provided. Required input when "
                         "running in 'field sheet and gui' mode! Try --help "
                         "option.")
            sys.exit(1)
        path2md = args.path2md
    else:
        if Path(args.path2md).is_dir():
            if qc_only:
                logger.info("Handling of the field sheets is bypassed when "
                            "running in 'quality control' mode.")
                path2md = None
            elif gui_only:
                logger.info("All metadata should be entered using the GUI in "
                            "'gui only' mode. Bypassing the use of field sheets.")
                path2md = None
            else:
                path2md = Path(args.path2md).resolve()
        else:
            logger.error("The path to the Phoenix field spreadsheets does not "
                         "exist!")
            if any([qc_only, gui_only]):
                logger.info("In 'quality control' or 'gui only' mode, field "
                            "sheets are not parsed.")
                path2md = None
            else:
                logger.error("A correct metadata path is a required input "
                             "when running in 'field sheet and gui' mode!")
                sys.exit(1)
    return path2md


def check_path2savedmetadata(args: argparse.Namespace, qc_only: bool, gui_only: bool) -> Optional[Path]:
    """Check path to saved metadata file(s)."""
    if args.path2savedmd is not None:
        if qc_only:
            logger.info("Handling of previously saved metadata is bypassed when "
                        "running in 'quality control' mode.")
            path2savedmd = None
        elif gui_only:
            if not Path(args.path2savedmd).is_file():
                logger.error("The provided path to a previously saved metadata "
                             "file (*.pkl) does not exist! Bypassing loading "
                             "of metadata fields before launching the GUI.")
                path2savedmd = None
            else:
                path2savedmd = Path(args.path2savedmd).resolve()
        else:
            logger.info("The loading of a previously saved metadata file (*.pkl) "
                        "is bypassed when running in 'field sheet and gui' "
                        "mode. Try --help option for more information.")
            path2savedmd = None
    else:
        path2savedmd = args.path2savedmd
    return path2savedmd


def check_output_dir(output_dir: Path, output_type: str) -> None:
    """
    Check whether the user-defined output directory exists.
    If not, create directory.
    """
    if not output_dir.exists():
        try:
            output_dir.mkdir(parents=True)
        except OSError:
            logger.exception("Failed to create output directory name for "
                             "{} files - {}".format(output_type, output_dir))
            sys.exit(1)


def get_args() -> Tuple[Path, Optional[Path], Optional[Path], bool, bool, Path, Optional[Path]]:
    """
    Parse command line arguments and perform a serie of checks on command line
    arguments.
    """
    # Set up command line interface and parse command line arguments
    parser = setup_args()
    args = parser.parse_args()

    # Check path to data directory, running modes, path to Phoenix field sheets
    # and path to .pkl file
    path2data = check_path2data(args)
    qc_only, gui_only = check_running_mode(args)
    path2md = check_path2metadata(args, qc_only, gui_only)
    path2savedmd = check_path2savedmetadata(args, qc_only, gui_only)

    # Check output directories
    output_mseed = Path(args.output_mseed).resolve()
    check_output_dir(output_mseed, 'day-long mseed')
    if qc_only:
        output_xml = None
    else:
        output_xml = Path(args.output_xml).resolve()
        check_output_dir(output_xml, 'StationXML')

    return path2data, path2md, path2savedmd, qc_only, gui_only, output_mseed, output_xml


def get_net_inputs() -> str:
    """Request user to enter 'network code'."""
    while True:
        net = input("Enter network code: ")
        try:
            valid = re.match(r"^\w{2}$", net)
        except TypeError:
            logger.error("The network code should be a string.")
            continue
        if valid is None:
            logger.error("The network code should be two alphanumeric "
                         "character long.")
            continue
        else:
            break
    return net


def get_sta_inputs() -> str:
    """Request user to enter 'station name'."""
    while True:
        sta = input("Enter station name: ")
        try:
            valid = re.match(r"^\w{3,5}$", sta)
        except TypeError:
            logger.error("The station name should be a string.")
            continue
        if valid is None:
            logger.error("The station name should be between three and five "
                         "alphanumeric character long.")
            continue
        else:
            break
    return sta


def get_num_inputs(sensor: str) -> int:
    """Request user to provide number of electrode pairs/coils installed."""
    valid_nums = ['0', '2'] if sensor == 'electrode pair' else ['2', '3']
    while True:
        try:
            nums = int(input("Enter number of {}s installed (valid options: {})."
                             .format(sensor, ', '.join(valid_nums))))
        except ValueError:
            logger.error("Please enter appropriate number of {}s (integer "
                         "expected)!".format(sensor))
            continue
        if nums not in [int(x) for x in valid_nums]:
            logger.error("Please enter appropriate number of {}s (valid "
                         "options: {})!".format(sensor, ', '.join(valid_nums)))
            continue
        else:
            break
    return nums


def get_cha_port_inputs(sensor: str, num: int, valid_cha_ports: List) -> str:
    """
    Request user to provide channel port to which each electrode pair/coil was
    connected.
    """
    while True:
        cha_port = input("Enter channel port for {} {} (valid options: {}): "
                         .format(sensor, num, ', '.join(valid_cha_ports)))
        if cha_port not in valid_cha_ports:
            logger.error("Please enter appropriate channel port number "
                         "(valid options: {})!".format(', '.join(valid_cha_ports)))
            continue
        else:
            break
    return cha_port


def get_dir_inputs(sensor: str, num: int, valid_dirs: Dict) -> int:
    """Request user to provide direction of each electrode pair/coil."""
    msg = ' & '.join(["{} for {}".format(k, v[0]) for k, v in valid_dirs.items()])
    while True:
        try:
            cha_dir = int(input("Enter direction of {} {} ({}): "
                                .format(sensor, num, msg)))
        except ValueError:
            logger.error("Please enter appropriate channel direction "
                         "(integer expected)!")
            continue
        if cha_dir not in valid_dirs.keys():
            logger.error("Please enter appropriate channel direction "
                         "(valid options: {})!".format(msg))
            continue
        else:
            break
    return cha_dir


def get_elec_mag_inputs(user_inputs: Dict, cha_type: str) -> Dict:
    """
    Request user to provide following iputs:
    - number of electrode pairs/coils installed
    - channel port to which each electrode pair/coil was connected (if any)
    - direction of each electrode pair/coil (if any)
    """
    if cha_type == 'elec':
        sensor = 'electrode pair'
        valid_cha_ports = [f'E{i+1}' for i in range(NUM_CHA_MAX['elec'])]
        valid_dirs = {1: ['North-South', 'Ex'], 2: ['East-West', 'Ey']}
    else:
        sensor = 'coil'
        valid_cha_ports = [f'H{i+1}' for i in range(NUM_CHA_MAX['mag'])]
        valid_dirs = {1: ['North-South', 'Hx'], 2: ['East-West', 'Hy'], 3: ['Vertical', 'Hz']}
    nums = get_num_inputs(sensor)
    for num in range(1, nums + 1):
        cha_port = get_cha_port_inputs(sensor, num, valid_cha_ports)
        user_inputs['chas_port'].append(cha_port)
        cha_dir = get_dir_inputs(sensor, num, valid_dirs)
        user_inputs['chas_info'][cha_port] = valid_dirs[cha_dir][1]
        valid_cha_ports.remove(cha_port)
        valid_dirs.pop(cha_dir)
    return user_inputs


def get_qc_inputs() -> Dict:
    """
    Request user inputs via the command line. Only use in QC mode.
    User inputs:
    - network code (key: net)
    - station name (key: sta)
    - number of electrode pairs installed
    - number of coils installed
    - channel port used for each electrode pair or coil (key: chas_port)
    - direction of each electrode pair and coil (key: chas_info)
    are stored in a dictionary.
    """
    logger.info("Please provide the following metadata before proceeding with "
                "the data conversion to day-long mseed files.")
    user_inputs: Dict[str, Any] = {'net': get_net_inputs(),
                                   'sta': get_sta_inputs(),
                                   'chas_port': [],
                                   'chas_info': {}}
    user_inputs = get_elec_mag_inputs(user_inputs, 'elec')
    user_inputs = get_elec_mag_inputs(user_inputs, 'mag')
    return user_inputs


def reformat_user_inputs(inputs: Dict) -> str:
    """
    Reformat qc input parameters in a user-friendly format. Reformatted inputs
    will be saved in the msg file and displayed in the command line.
    """
    direction = {'x': 'North-South', 'y': 'East-West', 'z': 'Vertical'}
    msg1 = (f"  Network code: {inputs['net']}\n"
            f"  Station name: {inputs['sta']}\n"
            f"  Channel ports used: {', '.join(inputs['chas_port'])}\n")
    msg2 = ""
    for cha, comp in inputs['chas_info'].items():
        msg2 += f"  Direction of channel connected to port {cha}: {direction[comp[-1]]}\n"
    msg = msg1 + msg2
    return msg.strip("\n")


def launch_gui(phoenix_data: PhoenixData, phoenix_md: PhoenixMetadata, path2savedmd: Optional[Path]) -> None:
    """
    Open GUI window, set the model data structure and handle metadata
    processing and archival steps.
    Only use in non-QC mode.
    """
    app = QtWidgets.QApplication(sys.argv)
    phoenix2seed_gui = PhoenixWindow(phoenix_data, phoenix_md)
    if path2savedmd:
        phoenix2seed_gui.load_md(path2savedmd)
    phoenix2seed_gui.show()
    sys.exit(app.exec_())


def main() -> None:
    # Read command line arguments
    path2data, path2md, path2savedmd, qc_only, gui_only, output_mseed, output_xml = get_args()

    # Parse data files
    phoenix_data = PhoenixData(path2data, output_mseed)
    phoenix_data.prep_data()

    # Directly convert recorded data to day-long mseed if QC mode was selected
    if qc_only:
        logger.info("Running phoenix2seed in 'quality control' mode.")
        qc_inputs = get_qc_inputs()
        logger.info("Summary of entered input parameters:\n{}"
                    .format(reformat_user_inputs(qc_inputs)))
        phoenix_data.update_data(qc_inputs=qc_inputs)
        logger.info("Writing day-long mseed files.")
        write_daylong_mseed(phoenix_data)
    # (Optional) Parse field sheets and prepopulate metadata in GUI
    # Populate metadata in GUI framework
    # Write day-long mseed and StationXML files after all required metadata
    # have been populated and validated.
    else:
        phoenix_md = PhoenixMetadata(path2md, output_xml, phoenix_data)  # type: ignore
        if not gui_only and not is_empty(phoenix_md.filenames):
            logger.info("Running phoenix2seed in 'field sheet and gui' mode.")
            md_fields = phoenix_md.from_field_sheets
            if md_fields is not None:
                logger.info("Pre-populating the GUI metadata fields using the "
                            "information provided in the field sheet(s).")
                phoenix_md.populate_props(md_fields)
        else:
            logger.info("Running phoenix2seed in 'gui only' mode.")
        logger.info("Launching the GUI.")
        # Launch GUI
        launch_gui(phoenix_data, phoenix_md, path2savedmd)


if __name__ == '__main__':
    main()
