# -*- coding: utf-8 -*-
"""
Routines for data prep before conversion to mseed files.

Maeva Pourpoint - IRIS/PASSCAL
"""

from __future__ import annotations

import importlib
import json
import pandas as pd
import re
import sys
import typing

from pathlib import Path
from statistics import mean
from typing import Dict, List, Optional, TYPE_CHECKING

from phoenix2seed.logging import setup_logger
from phoenix2seed.utils import (check_cha_id, get_run_list, gps_to_utc,
                                merge_nested_dict, VALID_EXTENSION)

if TYPE_CHECKING:
    from phoenix2seed.phoenix_metadata import PhoenixMetadata

# Set up logging
logger = setup_logger(__name__)

# Suggest new band code for channel with a sample rate >= 5000 Hz et corner
# period < 10 s: N
CHA_NAMING_CONV = {'Ex': {30: "SQN", 150: "EQN", 2400: "GQN", 24000: "NQN"},
                   'Ey': {30: "SQE", 150: "EQE", 2400: "GQE", 24000: "NQE"},
                   'Hx': {30: "SFN", 150: "EFN", 2400: "GFN", 24000: "NFN"},
                   'Hy': {30: "SFE", 150: "EFE", 2400: "GFE", 24000: "NFE"},
                   'Hz': {30: "SFZ", 150: "EFZ", 2400: "GFZ", 24000: "NFZ"}}


class PhoenixData:
    """
    This class handles the parsing and prep of the data outputted by MTU-5C.
    """

    def __init__(self, path2data: Path, output_mseed: Path) -> None:
        """Instantiate PhoenixData class."""
        self.path2data: Path = path2data
        self.output_mseed: Path = output_mseed
        self.num_runs: int = 0
        self.data_files: Dict = {}
        self.recmeta_files: List = []

    def order_recmeta_files(self) -> None:
        """Order recmeta files based on the time stamp recorded in their path."""
        self.recmeta_files.sort(key=lambda file: file.parts[-2].split('_')[-1])

    def scan_recmeta_files(self) -> None:
        """
        Scan user provided directory for Phoenix (MTU-5C) generated metadata
        files (recmeta.json).
        Example of metadata file path:
            .../NET/STA/AAAAA_YYYY-MM-DD-hhmmss/recmeta.json
            where NET is the network code and STA is the station name.
        Only the metadata file in the data acquisition folders should be
        scanned (i.e. folders containing channel_id folder (0, 1, 2, 3 and 4)).
        The other folders are receiver and sensor calibration folders.
        The number of recmeta files (i.e. number of data acquisition starts)
        defined the number of runs.
        """
        nbr_acq_dir = 0
        for path in self.path2data.glob('*'):
            if any([path.joinpath(x).is_dir() for x in ['0', '1', '2', '3', '4']]):
                nbr_acq_dir += 1
                recmeta_file = path.joinpath('recmeta.json')
                if recmeta_file.is_file():
                    self.recmeta_files.append(recmeta_file)
                    self.num_runs += 1
        if not self.recmeta_files:
            logger.error("No metadata file(s) (recmeta.json) found under the "
                         "following path - {}. This file should be "
                         "auto-generated by the loggger every time data "
                         "acquisition starts. Check data path!"
                         .format(self.path2data))
            sys.exit(1)
        if nbr_acq_dir != self.num_runs:
            logger.error("Number of metadata files (recmeta.json) found under "
                         "the following path - {} - is different than the "
                         "number of data acquisition folders. A metadata file "
                         "should be auto-generated by the loggger every time "
                         "data acquisition starts. One or more data "
                         "acquisition folders may have been corrupted!"
                         .format(self.path2data))
            sys.exit(1)
        self.order_recmeta_files()

    def set_stats(self) -> None:
        """Initialize stats attribute based on number of runs."""
        run_ids = get_run_list(self.num_runs)
        self.stats: Dict = {'status': {x: {} for x in run_ids},  # channel status (enabled or disabled)
                            'lp_filter': {x: {} for x in run_ids},  # low-pass filter in Hz
                            'coil_sn': {x: {} for x in run_ids},  # induction coil serial number
                            'contact_resistance_start': {x: {} for x in run_ids},  # contact resistance in Ω
                            'dipole_len': {x: {} for x in run_ids},  # dipole length between electrode pair in m
                            'decimation_scheme': {x: {} for x in run_ids},  # scheme used to decimate recorded data
                            'lat': None,  # latitude in decimal degree
                            'lon': None,  # longitude in decimal degree
                            'elev': None,  # elevation in m
                            'leap_seconds': {x: None for x in run_ids},  # leap seconds between GPS and UTC time in s
                            'cha_map': {x: {} for x in run_ids},  # Mapping between channel id and channek number
                            'datalogger_sn': {x: None for x in run_ids},  # receiver serial number
                            'acq_start_gps': {},  # acquisition start time in GPS
                            'acq_start_utc': {x: None for x in run_ids},  # acquisition start time for each run in UTC
                            'acq_end_utc': {x: None for x in run_ids},  # acquisition end time for each run in UTC
                            'start': None,  # start time of all acquisition runs
                            'end': None}  # end time of all acquisition runs

    def parse_recmeta_files(self) -> None:
        """
        Parse the recmeta.json files.
        Need to check with Phoenix that the layout of this file will remain
        unchanged between software updates.
        """
        run_ids = get_run_list(self.num_runs)
        latitude = []
        longitude = []
        elevation = []
        for ind, recmeta_file in enumerate(self.recmeta_files):
            with open(recmeta_file, 'r') as fin:
                meta = json.load(fin)
            status = {}
            lp = {}
            coil_sn = {}
            contact_resistance = {}
            dipole_len = {}
            for cha in meta['chconfig']['chans']:
                status[cha['tag']] = cha['on']
                lp[cha['tag']] = cha['lp']
                if cha.get('serial_detected'):  # field implemented in firmware version > 2.0
                    coil_sn[cha['tag']] = str(cha['serial_detected'])
                elif cha.get('serial'):
                    coil_sn[cha['tag']] = str(cha['serial'])
                if cha.get('pot_p'):
                    # Contact resistance measured with respect to the central "ground" electrode
                    contact_resistance[cha['tag']] = {'pot_p': cha['pot_p'],
                                                      'pot_n': cha['pot_n']}
                # Dipole length field has a default value of 50 m in recmeta.json file
                if cha.get('length1'):
                    dipole_len[cha['tag']] = cha['length1'] + cha['length2']
            self.stats['status'][run_ids[ind]] = status
            self.stats['lp_filter'][run_ids[ind]] = lp
            self.stats['coil_sn'][run_ids[ind]] = coil_sn
            self.stats['contact_resistance_start'][run_ids[ind]] = contact_resistance
            self.stats['dipole_len'][run_ids[ind]] = dipole_len
            self.stats['decimation_scheme'][run_ids[ind]] = {'interleave_0': meta["decimation"]['interleave_lev0'],  # in s  # noqa: E501
                                                             'interleave_1': meta["decimation"]['interleave_lev1'],  # in s  # noqa: E501
                                                             'period': meta["decimation"]['version_1']['period']}  # in s  # noqa: E501
            latitude.append(meta["timing"]["gps_lat"])
            longitude.append(meta["timing"]["gps_lon"])
            elevation.append(meta["timing"]["gps_alt"])
            self.stats['leap_seconds'][run_ids[ind]] = meta["timing"]["tm_leapSecAtRecStart"]
            self.stats['cha_map'][run_ids[ind]] = {str(x['idx']): x['tag'] for x in meta["channel_map"]["mapping"]}
            self.stats['datalogger_sn'][run_ids[ind]] = str(meta['instid'])
            self.stats['acq_start_gps'][meta['start']] = run_ids[ind]  # Unix-epoch based
        self.stats['lat'] = mean(latitude)
        self.stats['lon'] = mean(longitude)
        self.stats['elev'] = mean(elevation)

    @staticmethod
    def check_dir_structure(sd_dir: Path) -> bool:
        """Check data directory structure against expected structure."""
        if not re.search(r"\d{5}_\d{4}-\d{2}-\d{2}-\d{6}/\d$", str(sd_dir)):
            logger.warning("Unexpected directory structure! Skipping all data "
                           "files under the following directory {}."
                           .format(sd_dir))
            return False
        return True

    @staticmethod
    def check_file_extension(file: Path) -> bool:
        """Check data file extension against expected file extensions."""
        extension = file.suffix[1:]
        if extension not in VALID_EXTENSION.keys():
            logger.warning("Unexpected file extension! Skipping file {}."
                           .format(file))
            return False
        return True

    def order_data_files(self) -> None:
        """
        Order data files based on their index in the sequence of generated
        fragmented data files.
        """
        for key in self.data_files.keys():
            self.data_files[key].sort(key=lambda file: int(file.stem.split('_')[-1], 16))

    def scan_data_files(self) -> None:
        """
        Scan user provided directory for Phoenix (MTU-5C) generated data files.
        Examples of data file path:
          .../NET/STA/AAAAA_YYYY-MM-DD-hhmmss/C/AAAAA_BBBBBBBB_C_DDDDDDDD.bin
          .../NET/STA/AAAAA_YYYY-MM-DD-hhmmss/C/AAAAA_BBBBBBBB_C_DDDDDDDD.td_E
          where:
          - AAAAA is a 5 digit string indicating the serial number of the
            MTU-5C logger.
          - YYYY-MM-DD is the GPS date when data started being recorded.
          - hhmmss is the GPS time (UTC + leap seconds) when data started being
            recorded (hh are the hour in 24-hour format).
          - BBBBBBBB is a 32-bit unsigned hexadecimal label indicating the GPS
            timestamp of the start time of the recording.
          - C is a hexadecimal label indicating the channel id that the file
            corresponds to, starting at zero.
          - DDDDDDDD is a 32-bit unsigned hexadecimal label indicating the
            index of this file in the sequence of fragmented files.
          - E is a multi-character string that indicates the data sampling
            rate. Only applicable for decimated data. Current valid extensions:
            td_30, td_150, td_2400, td_24k.
        """
        for file in self.path2data.glob('**/*.*'):
            parent, name = file.parent, file.name
            if re.match(r"\d{5}_\w{8}_\d_\w{8}", name):
                if not PhoenixData.check_dir_structure(parent):
                    continue
                if not PhoenixData.check_file_extension(file):
                    continue
                if not check_cha_id(name, parent):
                    continue
                acq_start = int(name.split('_')[1], 16)
                cha_id = parent.parts[-1]
                srate_reader = '_'.join(VALID_EXTENSION[file.suffix[1:]])
                key = f"{acq_start}_{cha_id}_{srate_reader}"
                if self.data_files.get(key) is None:
                    self.data_files[key] = [file]
                else:
                    self.data_files[key].append(file)
        if not self.data_files:
            logger.error("No data files found under the following path - {}. "
                         "Check data path!".format(self.path2data))
            sys.exit(1)
        self.order_data_files()

    def parse_data_files(self) -> None:
        """
        Parse all data files and define data_list attribute => nested lists
        with following columns:
        - channel number (H1, H2, H3, E1 or E2)
        - component (Hx, Hy, Hz, Ex or Ey - Initialized to '')
        - channel name (See CHA_NAMING_CONV for mapping between component and
          channel_name - Initialized to '')
        - location code (Should be '' unless new suggested band code for sample
          rate > 5000 Hz and corner period < 10 s is not accepted. In which
          case, in the 24ksps longterm recording mode, two channels (2.4 and 24
          kHz) will have identical channel names. The following format
          [0-9][A-Z] will have to be used to differentiate between the two
          channels. Ex: 0A, 0B...)
        - run number (Number of runs is defined by the number of acquisition
          starts)
        - start time (Start time of data segment in UTC)
        - end time (End time of data segment in UTC)
        - sample rate (in Hz)
        - samples (np.array of data points)
        """
        self.data_list = []
        # Dynamically import phoenix_parser module
        module = importlib.import_module('phoenix2seed.phoenix_parser')
        for key in sorted(self.data_files.keys()):
            acq_start, cha_id, sample_rate, reader = key.split('_')
            run_id = self.stats['acq_start_gps'][int(acq_start)]
            cha_num = self.stats['cha_map'][run_id][cha_id]
            comp, cha_name, loc = ['', '', '']
            acq_start = gps_to_utc(int(acq_start), self.stats['leap_seconds'][run_id])
            logger.info("Parsing data files acquired on {} for channel {} "
                        "({}) and sample rate {} Hz."
                        .format(acq_start, cha_num, cha_id, sample_rate))
            # Load approptiate reader class based on reader's name
            class_ = getattr(module, reader)
            # Dynamic instantiation of reader class
            reader_inst = class_(self.stats['leap_seconds'][run_id])
            reader_inst.parse_data_files(self.data_files[key])
            if len(reader_inst.data) == 0:
                continue
            for samples, starttime in zip(reader_inst.data, reader_inst.time_stamps):
                sample_rate = int(sample_rate)
                endtime = starttime + (len(samples) - 1)/sample_rate
                self.data_list.append([cha_num, comp, cha_name, loc, run_id,
                                       starttime, endtime, sample_rate, samples])

    def create_dataframe(self) -> None:
        """Create pandas dataframe from parsed data."""
        columns = ['cha_num', 'comp', 'cha_name', 'loc', 'run_id',
                   'starttime', 'endtime', 'sample_rate', 'samples']
        self.data_frame = pd.DataFrame(self.data_list, columns=columns)

    def update_stats_time(self) -> None:
        """
        Update 'start', 'end', 'acq_start_utc' and 'acq_end_utc' stats values.
        The first two stats are used to validate user-provided metadata fields
        about the start and end time of their network and station.
        The last two stats are used to define the start and end time of each run.
        """
        run_ids = get_run_list(self.num_runs)
        for run_id in run_ids:
            ind = self.data_frame.run_id.eq(run_id)
            self.stats['acq_start_utc'][run_id] = self.data_frame[ind].starttime.min()
            self.stats['acq_end_utc'][run_id] = self.data_frame[ind].endtime.max()
        self.stats['start'] = self.stats['acq_start_utc'][run_ids[0]]
        self.stats['end'] = self.stats['acq_end_utc'][run_ids[-1]]

    def prep_data(self) -> None:
        """Wrap up method to prep data outputted by MTU-5C."""
        logger.info("Scanning directory - {} - for Phoenix (MTU-5C) generated "
                    "metadata files".format(self.path2data))
        self.scan_recmeta_files()
        self.set_stats()
        self.parse_recmeta_files()
        logger.info("Scanning directory - {} - for Phoenix (MTU-5C) generated "
                    "data files".format(self.path2data))
        self.scan_data_files()
        self.parse_data_files()
        if not self.data_list:
            logger.error('No valid data found. Exiting!')
            sys.exit(1)
        else:
            self.create_dataframe()
            self.update_stats_time()

    def update_stats_net_sta(self, net: str, sta: str) -> None:
        """
        Update stats attribute by defining or redefining the network code and
        station name.
        These stats info are used to set the header information for a ObsPy
        Trace object.
        """
        self.stats['net'] = net.upper()
        self.stats['sta'] = sta.upper()

    def get_non_rec_chas(self, run_id: str, user_record: List) -> List:
        """
        Check list of user-provided recorded channels against status of each
        channel (enabled vs. disabled).
        Returns list of channels for which data was not recorded (either
        because the channel was disabled or the user did not intend to record
        data for that channel but forgot to disable that channel when setting
        up the station).
        """
        status = self.stats['status'][run_id]
        status_record = [k for k, v in status.items() if v == 1]
        no_record = list(set(user_record) ^ set(status_record))
        return no_record

    def update_chas_rec(self, run_id: str, no_record: List) -> None:
        """Remove non-recorded channels from data frame."""
        for cha in no_record:
            ind = self.data_frame.cha_num.ne(cha) & self.data_frame.run_id.eq(run_id)
            self.data_frame = self.data_frame[ind]

    def update_chas_info(self, run_id: str, chas_info: Dict) -> None:
        """
        Update channel info (component and channel name) for given run in data frame.
        """
        for key, val in chas_info.items():
            ind_1 = self.data_frame.cha_num.eq(key) & self.data_frame.run_id.eq(run_id)
            self.data_frame.loc[ind_1, 'comp'] = val
            # Channel name depends on component and sampling rate
            for sample_rate in set(self.data_frame[ind_1].sample_rate):
                ind_2 = ind_1 & self.data_frame.sample_rate.eq(sample_rate)
                self.data_frame.loc[ind_2, 'cha_name'] = CHA_NAMING_CONV[val][sample_rate]

    def update_dataframe(self, chas_port: Dict, chas_info: Dict) -> None:
        """
        Update data frame based on:
        - list of channel ports used.
        - list of channels related info (mapping between channel port and
          component).
        """
        for run_id in get_run_list(self.num_runs):
            # Remove from data frame channels for which data were not recorded
            no_record = self.get_non_rec_chas(run_id, chas_port[run_id])
            self.update_chas_rec(run_id, no_record)
            # Update component and channel_name in data frame
            chas_info_run = {k: v for k, v in chas_info[run_id].items() if k not in no_record}
            self.update_chas_info(run_id, chas_info_run)

    @typing.no_type_check
    def update_data(self, qc_inputs: Optional[Dict] = None, md: Optional[PhoenixMetadata] = None) -> None:
        """
        Update stats and data_frame attributes based on user inputs (entered
        either via the command line or the GUI).
        Called either right after prep_data() in QC mode or after parsing
        metadata info in normal mode.
        """
        if qc_inputs is not None:
            net = qc_inputs['net']
            sta = qc_inputs['sta']
            run_list = get_run_list(self.num_runs)
            chas_info = {run_id: qc_inputs['chas_info'] for run_id in run_list}
            chas_port = {run_id: qc_inputs['chas_port'] for run_id in run_list}
        else:
            print('-------GUI-------')
            net = md.net.archive_net
            sta = md.sta.archive_id
            e_infos = md.get_cha_infos('elec')
            m_infos = md.get_cha_infos('mag')
            chas_info = merge_nested_dict(e_infos, m_infos)
            chas_port = {k: list(v.keys()) for k, v in chas_info.items()}
        self.update_stats_net_sta(net, sta)
        self.update_dataframe(chas_port, chas_info)
