# -*- coding: utf-8 -*-
"""
Routines for parsing Phoenix field spreadsheets (i.e. install and credential
sheets) and populating metadata fields;
Field sheets are provided to the PIs and their formatting is hard coded.

Maeva Pourpoint - IRIS/PASSCAL
"""

from __future__ import annotations

import copy
import openpyxl
import pickle

from dataclasses import fields
from datetime import datetime
from inspect import signature
from obspy import UTCDateTime
from openpyxl.worksheet.worksheet import Worksheet
from operator import methodcaller
from pathlib import Path
from typing import Dict, List, Optional, TYPE_CHECKING, Union

from phoenix2seed.logging import parse_config_ini, setup_logger
from phoenix2seed.metadata_category import Elec, Mag, Run, Sta, Net, MSG_MAP
from phoenix2seed.utils import (is_empty, get_ids, get_run_list, NUM_CHA_MAX,
                                str2list)

if TYPE_CHECKING:
    import pandas as pd
    from phoenix2seed.phoenix_data import PhoenixData

# Read config.ini file and set up logging
config = parse_config_ini()
logger = setup_logger(__name__)

# -- Type aliases --
MD_VAL = Union[str, int, float, datetime]  # Type of metadata value
DCS = Union[Net, Sta, Run, Elec, Mag]  # Type of dataclasses
DCS_SUB_REM = Union[Run, Elec, Mag]
DCS_SUB_EM = Union[Elec, Mag]
# -- Type aliases --

IDS = {'elec': {'cha_id': 'E', 'sensor': 'electrode pair', 'method': Elec,
                'level': 'electric'},
       'mag': {'cha_id': 'H', 'sensor': 'coil', 'method': Mag,
               'level': 'magnetic'}}


class PhoenixMetadata:
    """
    This class handles the parsing of the Phoenix field sheets and the
    population of the metadata fields following validation.
    """

    def __init__(self, path2md: Path, output_xml: Path, phoenix_data: PhoenixData) -> None:
        """Instantiate PhoenixMetadata class and scan metadata directory."""
        self.path2md = path2md
        self.output_xml = output_xml
        self.filenames: List[Path] = []
        self.cats: List[str] = ['Net', 'Sta', 'Run', 'Elec', 'Mag']
        self.net: Net = Net()
        self.sta: Sta = Sta()
        self.run: List[Run] = []
        self.elec: List[Elec] = []
        self.mag: List[Mag] = []
        self.data_stats: Dict = phoenix_data.stats
        self.data_frame: pd.DataFrame = phoenix_data.data_frame
        self.num_runs = phoenix_data.num_runs
        self.run_list: List[str] = get_run_list(self.num_runs)
        if path2md is not None:
            self.scan_path2md()

    def scan_path2md(self) -> None:
        """
        Scan metadata directory. If no field sheets are found under the given
        directory, the user will have to provide the necessary metadata in the
        GUI.
        """
        filenames = [tmp for tmp in self.path2md.glob('*.xlsx')
                     if not tmp.parts[-1].startswith('~$')]
        if not filenames:
            logger.error("No field sheet found under the following path - {}. "
                         "All metadata will have to be provided using the "
                         "GUI!".format(self.path2md))
        else:
            self.filenames = filenames

    @staticmethod
    def get_md_field(sheet: Worksheet, cell: str) -> Optional[MD_VAL]:
        """Get metadata field at a given cell in a given field sheet."""
        try:
            field_val = sheet[cell].value
        except Exception as error:
            logger.exception("Failed to retrieve value of cell {} - Skipping "
                             "cell - Exception: {}".format(cell, error))
            return None
        return field_val

    @staticmethod
    def identify_field_sheet(sheet: Worksheet, filename: Path) -> Optional[str]:
        """
        Check whether a field sheet is an installation or credential sheet.
        Layout between provided spreadsheet templates varies and cell numbers
        listed in config file are organized in two sub-categories:
        install_sheet and credential_sheet.
        """
        sheet_header = PhoenixMetadata.get_md_field(sheet, config['sheets']['header'])
        sheet_types = {val: key for key, val in config['sheets'].items()[1:]}
        sheet_type = sheet_types.get(sheet_header)
        if sheet_type is None:
            logger.warning("The following file {} does not have the proper "
                           "header. The provided spreadsheet template was "
                           "not used or its layout was modified. Skipping "
                           "file!".format(filename))
        return sheet_type

    def init_run_props(self) -> List[Run]:
        """
        Instantiate Run data class object and initialize non-user defined
        properties.
        """
        runs = []
        for run_id in self.run_list:
            ind = self.data_frame.run_id.eq(run_id)
            sample_rates = sorted(set(self.data_frame[ind].sample_rate))
            run = Run(resource_id=f'mt.run.id:{run_id}',
                      start=self.data_stats['acq_start_utc'][run_id],
                      end=self.data_stats['acq_end_utc'][run_id],
                      sample_rates=' - '.join(str(x) for x in sample_rates))
            runs.append(run)
        return runs

    def init_cha_props(self, cha_type: str, run_id: str) -> DCS_SUB_EM:
        """
        Create instances of Elec or Mag data class and initialize properties
        inherited from the BaseSta and BaseChannel data classes.
        """
        ChaClass = globals()[cha_type.capitalize()]
        cha = ChaClass(elev=round(self.data_stats["elev"], 3),
                       lat=round(self.data_stats["lat"], 3),
                       lon=round(self.data_stats["lon"], 3),
                       run_id=run_id)
        return cha

    def filter_cha(self, cha_type: str, run_id: str) -> List[DCS_SUB_EM]:
        """
        For a given channel type (electric or magnetic), get list of channels
        with a given run id.
        """
        return [c for c in getattr(self, cha_type) if c.run_id == run_id]

    @staticmethod
    def update_cha(cha: DCS_SUB_EM, cha_port: str) -> DCS_SUB_EM:
        """
        Make a copy of the populated Elec or Mag data class and set its channel
        port and electrode pair/coil number.
        """
        # Use deepcopy here so that any changes made to the copy object do not
        # reflect in the original object.
        tmp = copy.deepcopy(cha)
        tmp.cha_port = cha_port
        tmp.cha_num = int(cha_port[1])
        return tmp

    @staticmethod
    def dc2dict(dc: DCS_SUB_REM) -> Dict:
        """
        Convert dataclass instance to a dict and filter dict to:
        - only keep metadata fields displayed in GUI
        - reset missing and invalid metadata fields
        """
        missing_invalid = dc.md_missing | dc.md_invalid
        mdict = {}
        for md_field in fields(dc):
            if md_field.metadata.get('gui'):
                if md_field.name not in missing_invalid:
                    mdict[md_field.name] = getattr(dc, md_field.name)
                else:
                    mdict[md_field.name] = md_field.default
        return mdict

    def get_chas_dict(self, cat: str, output: Dict) -> Dict:
        """
        Get dictionary representation of all electric or magnetic data classes.
        """
        cha_type = cat.lower()
        cha_id = IDS[cha_type]['cha_id']
        for run_id in self.run_list:
            chas = self.filter_cha(cha_type, run_id)
            if not chas:
                cha = self.init_cha_props(cha_type, run_id)
                chas = [PhoenixMetadata.update_cha(cha, f'{cha_id}{ind + 1}')
                        for ind in range(NUM_CHA_MAX[cha_type])]
            for ind in range(NUM_CHA_MAX[cha_type]):
                key = f'Run_{run_id}_{cha_type}_{ind+1}'
                output[cat][key] = PhoenixMetadata.dc2dict(chas[ind])
        return output

    @property
    def for_gui(self) -> Dict:
        """
        Define metadata output dictionary structure that will be used to set
        the model in the GUI.
        """
        output: Dict = {}
        for cat in self.cats:
            output[cat] = {}
            mprop = getattr(self, cat.lower())
            if isinstance(mprop, list):
                if cat == 'Run':
                    if not mprop:
                        mprop = self.init_run_props()
                    for ind, dc in enumerate(mprop):
                        key = f'Run_{self.run_list[ind]}'
                        output[cat][key] = PhoenixMetadata.dc2dict(dc)
                else:
                    output = self.get_chas_dict(cat, output)
            else:
                output[cat] = PhoenixMetadata.dc2dict(mprop)
        return output

    def parse_field_sheet(self, filename: Path, sheet: Worksheet, sheet_type: str, md_fields: Dict) -> Dict:
        """
        Parse field sheet using the cell numbers listed in the config file and
        return a nested dictionary of metadata entries organized based on the
        category under which a given metadata field falls.
        """
        logger.info("Parsing the {} {} - {}.".format(*sheet_type.split('_'), filename))
        for cat in self.cats:
            if cat not in md_fields:
                md_fields[cat] = {}
            if sheet_type not in config[cat]:
                continue
            logger.info("Parsing metadata fields at the {} level.".format(MSG_MAP[cat]))
            for key, val in config[cat][sheet_type].items():
                if isinstance(val, list):
                    field_val = [PhoenixMetadata.get_md_field(sheet, x) for x in val]
                else:
                    field_val = PhoenixMetadata.get_md_field(sheet, val)  # type: ignore
                if key == "comments":
                    field_val = ', '.join(filter(None, (md_fields[cat].get(key, None), field_val)))  # type: ignore
                if key == 'dipole_len':
                    field_val = [sum(map(abs, field_val[0:2])), sum(map(abs, field_val[2:]))]  # type: ignore
                md_fields[cat][key] = field_val
        return md_fields

    def reformat_run(self, run_fields: Dict) -> Dict:
        """Reformat run metadata fields."""
        tmp = {f'Run_{run_id}': dict.fromkeys(run_fields.keys()) for run_id in self.run_list}
        tmp['Run_a'].update(run_fields)
        return tmp

    def reformat_cha(self, cha_type: str, cha_fields: Dict) -> Dict:
        """Reformat metadata fields for electric or magnetic channel."""
        ids = get_ids(self.num_runs, cha_type)
        tmp = {f'Run_{run_id}_{cha_type}_{ind}': dict.fromkeys(cha_fields.keys())
               for run_id, ind in ids}
        for key, val in cha_fields.items():
            for ind in range(1, NUM_CHA_MAX[cha_type] + 1):
                if isinstance(val, list):
                    cha_val = val[ind - 1]
                else:
                    cha_val = val
                tmp[f'Run_a_{cha_type}_{ind}'][key] = cha_val
        return tmp

    def reformat_md_dict(self, md_fields: Dict) -> Dict:
        """Reformat run, electric and magnetic metadata_fields."""
        md_fields['Run'] = self.reformat_run(md_fields['Run'])
        md_fields['Elec'] = self.reformat_cha('elec', md_fields['Elec'])
        md_fields['Mag'] = self.reformat_cha('mag', md_fields['Mag'])
        return md_fields

    @staticmethod
    def check_field_sheets(sheet_types: List[Optional[str]]) -> None:
        """
        Check that all valid sheet types have been found and parsed. If not,
        warn user of missing sheet(s).
        """
        valid_types = [key for key in config['sheets'].keys()[1:]]
        for valid_type in valid_types:
            if sheet_types and valid_type not in sheet_types:
                logger.warning("No valid {0} sheet found. The {0} metadata "
                               "fields will have to be provided using the GUI."
                               .format(valid_type.split('_')[0]))

    @property
    def from_field_sheets(self) -> Optional[Dict]:
        """Parse all field sheets provided by the user."""
        md_fields: Dict = {}
        sheet_types = []
        parsed_sheets = []
        for filename in self.filenames:
            try:
                workbook = openpyxl.load_workbook(filename, data_only=True)  # type: ignore
            except Exception as error:
                logger.error("Failed to open field sheet: {} - Skipping file "
                             "- Exception: {}".format(filename, error))
            else:
                if workbook is None:
                    continue
                sheet = workbook.active
                sheet_type = PhoenixMetadata.identify_field_sheet(sheet, filename)
                sheet_types.append(sheet_type)
                if sheet_type is None:
                    continue
                if sheet_type not in parsed_sheets:
                    parsed_sheets.append(sheet_type)
                else:
                    logger.warning("Already parsed the {0} {1} - You may have "
                                   "more than one {0} {1} - Skipping {2}!"
                                   .format(*sheet_type.split('_'), filename))
                    continue
                md_fields = self.parse_field_sheet(filename, sheet, sheet_type, md_fields)
                workbook.close()
        PhoenixMetadata.check_field_sheets(sheet_types)
        return self.reformat_md_dict(md_fields) if not is_empty(md_fields) else None

    def populate(self, cat: DCS, md_fields: Dict, run_id: Optional[str] = None) -> None:
        """
        Populate metadata properties for given "category" of dataclass.
        Loop over all metadata fields for given "category" and call appropriate
        validation method based on metadata fields name.
        If validation method exists, use method to validate metadata field value.
        Some metadata inputs are validated against data inputs. Ex: latitude,
        longitude, elevation, serial numbers, dipole length ...
        """
        for key, val in md_fields.items():
            valid = True
            validate_method = f'validate_{key}'
            if isinstance(val, datetime):
                val = UTCDateTime(val)
            if hasattr(cat, validate_method):
                if "data_input" in signature(getattr(cat, validate_method)).parameters:
                    data_input = self.data_stats[key]
                    if isinstance(data_input, dict):
                        data_input = self.data_stats[key][run_id]
                    valid = methodcaller(validate_method, val, data_input)(cat)
                else:
                    valid = methodcaller(validate_method, val)(cat)
            if valid:
                setattr(cat, key, val)
                cat.md_invalid.discard(key)
            else:
                cat.md_invalid.add(key)

    @staticmethod
    def flag_md_missing(cat: DCS, skip: List[str] = []) -> None:
        """
        Flag metadata fields that are required for archiving but "missing"
        because not provided by the user.
        """
        md_props = cat.__dict__
        md_req = [md_field.name for md_field in fields(cat) if md_field.metadata.get('req')]
        for key, val in md_props.items():
            if key in md_req and key not in skip:
                if is_empty(val):
                    cat.md_missing.add(key)
                else:
                    cat.md_missing.discard(key)

    def populate_net_props(self, md_fields_n: Dict) -> None:
        """
        Populate properties of Net data class based on user inputs from the
        field sheets and/or GUI.
        """
        self.populate(self.net, md_fields_n)
        PhoenixMetadata.flag_md_missing(self.net)

    def check_sta_start_end(self) -> bool:
        """
        Check that the start and end times of a station fall within the
        network start and end times.
        """
        reset = False
        for time_field in ['start', 'end']:
            time_net = getattr(self.net, time_field)
            if time_field in self.sta.md_invalid or time_net is None:
                continue
            time_sta = getattr(self.sta, time_field)
            valid = time_sta >= time_net if time_field == "start" else time_sta <= time_net
            if not valid:
                msg = "greater" if time_field == "start" else "lower"
                logger.error("The {0} date of the station should be {1} than the {0} date "
                             "of the network.".format(time_field, msg))
                logger.warning("Setting the {0} date of the station to be the same as the "
                               "{0} date of the network.".format(time_field))
                setattr(self.sta, time_field, time_net)
                reset = True
        return reset

    def populate_sta_props(self, md_fields_s: Dict) -> bool:
        """
        Populate properties of Sta data class based on user inputs from the
        field sheets and/or GUI.
        """
        self.populate(self.sta, md_fields_s)
        self.sta.run_list = ', '.join(self.run_list)
        reset = self.check_sta_start_end()
        PhoenixMetadata.flag_md_missing(self.sta)
        return reset

    def get_run(self, run_id: str) -> Run:
        """Get Run data class for given run id."""
        ind_run = self.run_list.index(run_id)
        return self.run[ind_run]

    def populate_run_props(self, md_fields_r: Dict, run_list: Optional[List] = None) -> None:
        """
        Populate properties of Run data class based on user inputs from the
        field sheets and/or GUI.
        """
        run_list_ = self.run_list if run_list is None else run_list
        if not self.run:
            self.run.extend(self.init_run_props())
        for run_id in run_list_:
            run = self.get_run(run_id)
            name = f'Run_{run_id}'
            self.populate(run, md_fields_r[name], run_id)
            PhoenixMetadata.flag_md_missing(run)

    def get_chas_rec(self, cha_id: str, run_id: str) -> list:
        """
        For a given channel type (electric or magnetic) and a given run_id,
        get list of channels recorded.
        """
        run = self.get_run(run_id)
        if run.chas_rec is None:
            return []
        else:
            return [x for x in str2list(run.chas_rec, sep='-') if x.startswith(cha_id)]  # type: ignore

    def match_num_chas(self, cha_type: str, chas_rec: List[str], run_id: str, num_chas: int) -> None:
        """
        Check that number of electrode pairs/coils specified at the Elec/Mag
        level for a given run matches the number of electric/magnetic channels
        specified by the user at the run level.
        """
        sensor = IDS[cha_type]['sensor']
        cha_level = IDS[cha_type]['level']
        chas = self.filter_cha(cha_type, run_id)
        if num_chas != len(chas_rec):
            logger.error("Invalid number of {0}s (run '{1}')! The number of {0}s "
                         "does not match the number of {2} channels in your "
                         "list of recorded channels at the Station/Run level."
                         .format(sensor, run_id, cha_level))
            for cha in chas:
                cha.md_invalid.add(f'run_{run_id}_num_{cha_type}')
        else:
            for cha in chas:
                cha.md_invalid.discard(f'run_{run_id}_num_{cha_type}')

    def get_cha_infos(self, cha_type: str) -> Dict:
        """
        Get electrode pair/coil information (channel port and associated component)
        for each run.
        """
        cha_id = IDS[cha_type]['cha_id']
        infos = {}
        for run in self.run:
            run_id = run.run_id
            chas_rec = self.get_chas_rec(cha_id, run_id)  # type: ignore
            cha = self.filter_cha(cha_type, run_id)
            infos[run_id] = {x.cha_port: x.comp for x in cha if f'{cha_id}{x.cha_num}' in chas_rec}
        return infos

    def populate_chas_props(self, cha_type: str, md_fields: Dict, run_list: Optional[List] = None,
                            num_chas: Optional[Dict] = None) -> Dict:
        """
        Populate properties of channel data class (Elec or Mag) based on user
        inputs from the field sheets and/or GUI.
        """
        cha_id = IDS[cha_type]['cha_id']
        cha_method = IDS[cha_type]['method']
        # Get list of set methods for Elec or Mag data class
        set_methods = [x for x in dir(cha_method) if x.startswith('set_')]
        # Get complete list of runs or a subset of it
        run_list_ = self.run_list if run_list is None else run_list
        num_chas_ = {}
        for run_id in run_list_:
            # Get list of electric or magnetic channels recorded for given run id
            chas_rec = self.get_chas_rec(cha_id, run_id)  # type: ignore
            # Skip populating the Elec/Mag data class if no channel recorded
            if not chas_rec:
                continue
            # Create instances of Elec/Mag data class and initialize properties
            if not self.filter_cha(cha_type, run_id):
                cha = self.init_cha_props(cha_type, run_id)
                chas = getattr(self, cha_type)
                chas.extend([PhoenixMetadata.update_cha(cha, f'{cha_id}{i+1}')
                             for i in range(NUM_CHA_MAX[cha_type])])
            # Get number of electrode pairs/coils from user input
            num_chas_[run_id] = (int(num_chas.get(run_id) or len(chas_rec))
                                 if num_chas else len(chas_rec))
            # Populate/Validata metadata fields for all recorded electric or
            # magnetic channels for a given run id
            for ind, cha in enumerate(self.filter_cha(cha_type, run_id)):
                key = f'Run_{run_id}_{cha_type}_{ind+1}'
                if ind in range(num_chas_[run_id]):
                    self.populate(cha, md_fields[key], run_id)
                    for set_method in set_methods:
                        methodcaller(set_method)(cha)
                    PhoenixMetadata.flag_md_missing(cha)
                else:
                    cha.cha_port = md_fields[key]['cha_port']
            self.match_num_chas(cha_type, chas_rec, run_id, num_chas_[run_id])
        return num_chas_

    def populate_props(self, md_fields: Dict) -> None:
        """
        Populate metadata properties based on user inputs from the field sheets
        and/or GUI.
        """
        for cat in self.cats:
            md_fields_ = md_fields.get(cat)
            if cat in ['Elec', 'Mag']:
                populate_method = 'populate_chas_props'
                methodcaller(populate_method, cat.lower(), md_fields_)(self)
            else:
                populate_method = f'populate_{cat.lower()}_props'
                methodcaller(populate_method, md_fields_)(self)

    def save_md(self, filename: str) -> None:
        """
        Save instance of PhoenixMetadata class into a byte stream.
        Useful if:
        - the user wants to update metadata fields and regenerate StationXML
          files after terminating phoenix2seed.
        - the user wants to reuse some metadata fields for another
          station.
        - phoenix2seed is terminated unexpectedly.
        """
        with open(filename, 'wb') as fout:
            logger.info("Saving metadata inputs in {}".format(filename))
            pickle.dump(self, fout)

    @staticmethod
    def load_md(filename: str) -> PhoenixMetadata:
        """Load saved instance of PhoenixMetadata class."""
        with open(filename, 'rb') as fin:
            logger.info("Loading metadata inputs from {}".format(filename))
            phoenix_md = pickle.load(fin)
        return phoenix_md
