# -*- coding: utf-8 -*-
"""
Module to read and parse native Phoenix Geophysics data formats of the MTU-5C
Family.

This module implements Streamed readers for:
- continuous-decimated;
- segmented-decimated;
- native sampling rate;
time series formats of the MTU-5C family.

Authored by: Jorge Torres-Solis (Phoenix).
Refactored by: Maeva Pourpoint (IRIS/PASSCAL).
"""

import string
import sys

from numpy import append, empty, fromfile, float32, ndarray
from obspy import UTCDateTime
from pathlib import Path
from struct import unpack_from, unpack
from typing import Any, Dict, List, Tuple

from phoenix2seed.logging import setup_logger
from phoenix2seed.utils import check_cha_id, check_sample_rate, gps_to_utc

# Set up logging
logger = setup_logger(__name__)


class DataScaling:
    """
    This class contains the different data scaling types that can be produced
    by a time series reader of the MTU-5C Family.
    """
    AD_in_ADunits, AD_input_volts, instrument_input_volts = range(3)


class _TSReaderBase:
    def __init__(self, leap_seconds: int, header_size: int) -> None:
        """
        Instantiate _TSReaderBase class.
        Inherited by: DecimatedContinuousReader, DecimatedSegmentedReader and
        NativeReader classes.
        """
        self.leap_seconds: int = leap_seconds
        self.header_size: int = header_size
        self.header_info: Dict[str, Any] = {}
        self.master_header: bytes = b''
        self.data: List[ndarray] = []
        self.time_stamps: List[UTCDateTime] = []

    def __populate_channel_type(self, config_fp: Tuple) -> None:
        """Set channel type (E for electric & H for magnetic)."""
        if config_fp[1] & 0x08 == 0x08:
            self.channel_type = "E"
        else:
            self.channel_type = "H"
        # Channel type detected by electronics.
        # This normally matches self.channel_type, but used in electronics
        # design and testing
        if config_fp[1] & 0x20 == 0x20:
            self.detected_channel_type = 'E'
        else:
            self.detected_channel_type = 'H'

    def __populate_preamp_gain(self, config_fp: Tuple) -> None:
        """
        Set preamplifier gain. Gain of 1 for magnetic channel and 4 or 8 for
        electric channel if preamplifier is on.
        """
        preamp_on = bool(config_fp[0] & 0x10)
        self.preamp_gain = 1.0
        if self.channel_type == "E":
            if preamp_on is True:
                if self.board_model_main == "BCM01" or self.board_model_main == "BCM03":
                    self.preamp_gain = 4.0
                    if self.board_model_revision == "L":
                        # Account for BCM01-L experimental prototype
                        self.preamp_gain = 8.0
                else:
                    self.preamp_gain = 8.0
                    # Acount for experimental prototype BCM05-A
                    if self.header_info['ch_hwv'][0:7] == "BCM05-A":
                        self.preamp_gain = 4.0

    def __populate_lpf(self, config_fp: Tuple) -> None:
        """
        Set the nominal cutoff frequency of the configured LPF of the channel.
        """
        if config_fp[0] & 0x80 == 0x80:  # LPF on
            if config_fp[0] & 0x03 == 0x03:
                self.lpf_Hz = 10
            elif config_fp[0] & 0x03 == 0x02:
                if self.board_model_main == "BCM03" or self.board_model_main == "BCM06":
                    self.lpf_Hz = 1000
                else:
                    self.lpf_Hz = 100
            elif config_fp[0] & 0x03 == 0x01:
                if self.board_model_main == "BCM03" or self.board_model_main == "BCM06":
                    self.lpf_Hz = 10000
                else:
                    self.lpf_Hz = 1000
        else:  # LPF off
            if self.board_model_main == "BCM03" or self.board_model_main == "BCM06":
                self.lpf_Hz = 17800
            else:
                self.lpf_Hz = 10000

    def __populate_main_gain(self, config_fp: Tuple) -> None:
        """
        Set the value of the main gain of the board.
        BCM05-B and BCM06 introduced different selectable gains.
        We assume that any newer boards will have the new gain banks.
        """
        new_gains = True
        if self.board_model_main == "BCM01" or self.board_model_main == "BCM03":
            # Original style 24 KSps boards and original 96 KSps boards
            new_gains = False
        if self.header_info['ch_hwv'][0:7] == "BCM05-A":
            # Acount for experimental prototype BCM05-A, which also had original
            # gain banks
            new_gains = False
        if config_fp[0] & 0x0C == 0x00:
            self.channel_main_gain = 1.0
        elif config_fp[0] & 0x0C == 0x04:
            self.channel_main_gain = 4.0
        elif config_fp[0] & 0x0C == 0x08:
            self.channel_main_gain = 6.0
            if not new_gains:
                self.channel_main_gain = 16.0
        elif config_fp[0] & 0x0C == 0x0C:
            self.channel_main_gain = 8.0
            if not new_gains:
                self.channel_main_gain = 32.0

    def __handle_sensor_range(self, config_fp: Tuple) -> None:
        """
        This function will adjust the intrinsic circuitry gain based on the
        sensor range configuration in the configuration fingerprint.
        The intrinsic circuitry gain cannot directly be configured by the user.

        For this, we consider that for the Electric channel, calibration path,
        or H-legacy sensors all go through a 1/4 gain stage, and then they get
        a virtual x2 gain from Single-ended-diff before the A/D.
        In the case of newer sensors (differential) instead of a 1/4 gain stage,
        there is only a 1/2 gain stage.
        Therefore, in the Electric channel and legacy sensor case, the circuitry
        gain is 1/2 while for newer sensors it is 1.
        """
        self.intrinsic_circuitry_gain = 0.5
        if self.channel_type == "H":
            if config_fp[1] & 0x01 == 0x01:
                self.intrinsic_circuitry_gain = 1.0

    def __populate_attenuator_gain(self, config_fp: Tuple) -> None:
        """Set the attenuator gain value."""
        self.attenuator_gain = 1.0  # Assuming that the attenuator is off
        attenuator_on = bool(config_fp[4] & 0x01)
        if attenuator_on and self.channel_type == "E":
            # By default, assume that we are dealing with a newer type of boards
            new_attenuator = True
            if self.board_model_main == "BCM01" or self.board_model_main == "BCM03":
                # Original style 24 KSps boards and original 96 KSps boards
                new_attenuator = False
            if self.header_info['ch_hwv'][0:7] == "BCM05-A":
                # Acount for experimental prototype BCM05-A, which also had
                # original gain banks
                new_attenuator = False
            if new_attenuator:
                self.attenuator_gain = 523.0 / 5223.0
            else:
                self.attenuator_gain = 0.1

    def unpack_header(self) -> None:
        """
        Unpack Master Header fields.
        Fields size, byte offset, type and value are summarized in Tables 1 and
        2 in DAA09 doc under Docs folder.
        """
        self.header_info['file_type'] = unpack_from('B', self.master_header, 0)[0]
        self.header_info['file_version'] = unpack_from('B', self.master_header, 1)[0]
        self.header_info['length'] = unpack_from('H', self.master_header, 2)[0]
        self.header_info['inst_type'] = unpack_from('8s', self.master_header, 4)[0].decode("utf-8").strip(' ').strip('\x00')  # noqa: E501
        self.header_info['inst_serial'] = b''.join(unpack_from('cccccccc', self.master_header, 12)).strip(b'\x00')
        self.header_info['rec_id'] = unpack_from('I', self.master_header, 20)[0]
        self.header_info['ch_id'] = unpack_from('B', self.master_header, 24)[0]
        self.header_info['file_sequence'] = unpack_from('I', self.master_header, 25)[0]
        self.header_info['frag_period'] = unpack_from('H', self.master_header, 29)[0]
        self.header_info['ch_hwv'] = unpack_from('8s', self.master_header, 31)[0].decode("utf-8").strip(' ')
        self.board_model_main = self.header_info['ch_hwv'][0:5]
        self.board_model_revision = self.header_info['ch_hwv'][6:7]
        self.header_info['ch_ser'] = unpack_from('8s', self.master_header, 39)[0].decode("utf-8").strip('\x00')
        # handle the case of backend < v0.14, which puts '--------' in ch_ser
        if all(chars in string.hexdigits for chars in self.header_info['ch_ser']):
            self.header_info['ch_ser'] = int(self.header_info['ch_ser'], 16)
        else:
            self.header_info['ch_ser'] = 0
        self.header_info['ch_fir'] = hex(unpack_from('I', self.master_header, 47)[0])
        config_fp = unpack_from('BBBBBBBB', self.master_header, 51)
        self.header_info['conf_fp'] = config_fp
        # Channel type
        self.__populate_channel_type(config_fp)
        # Electric channel preamp
        self.__populate_preamp_gain(config_fp)
        # LPF
        self.__populate_lpf(config_fp)
        # Main gain stage
        self.__populate_main_gain(config_fp)
        # Sensor range
        self.__handle_sensor_range(config_fp)
        # Electric channel attenuator
        self.__populate_attenuator_gain(config_fp)
        # Board-wide gains
        self.total_selectable_gain = self.channel_main_gain * self.preamp_gain * self.attenuator_gain
        self.total_circuitry_gain = self.total_selectable_gain * self.intrinsic_circuitry_gain
        self.header_info['sample_rate_base'] = unpack_from('H', self.master_header, 59)[0]
        self.header_info['sample_rate_exp'] = unpack_from('b', self.master_header, 61)[0]
        self.header_info['sample_rate'] = self.header_info['sample_rate_base']
        if self.header_info['sample_rate_exp'] != 0:
            self.header_info['sample_rate'] *= pow(10, self.header_info['sample_rate_exp'])
        self.header_info['bytes_per_sample'] = unpack_from('B', self.master_header, 62)[0]
        self.header_info['frame_size'] = unpack_from('I', self.master_header, 63)[0]
        self.dataFooter = self.header_info['frame_size'] >> 24
        self.frameSize = self.header_info['frame_size'] & 0x0ffffff
        self.header_info['decimation_node_id'] = unpack_from('H', self.master_header, 67)[0]
        self.header_info['frame_rollover_count'] = unpack_from('H', self.master_header, 69)[0]
        self.header_info['gps_long'] = unpack_from('f', self.master_header, 71)[0]
        self.header_info['gps_lat'] = unpack_from('f', self.master_header, 75)[0]
        self.header_info['gps_height'] = unpack_from('f', self.master_header, 79)[0]
        self.header_info['gps_hacc'] = unpack_from('I', self.master_header, 83)[0]
        self.header_info['gps_vacc'] = unpack_from('I', self.master_header, 87)[0]
        self.header_info['timing_status'] = unpack_from('BBH', self.master_header, 91)
        self.header_info['timing_flags'] = self.header_info['timing_status'][0]
        self.header_info['timing_sat_count'] = self.header_info['timing_status'][1]
        self.header_info['timing_stability'] = self.header_info['timing_status'][2]
        self.header_info['future1'] = unpack_from('b', self.master_header, 95)[0]
        self.header_info['future2'] = unpack_from('i', self.master_header, 97)[0]
        self.header_info['saturated_frames'] = unpack_from('H', self.master_header, 101)[0]
        if self.header_info['saturated_frames'] & 0x80 == 0x80:
            self.header_info['saturated_frames'] &= 0x7F
            self.header_info['saturated_frames'] <<= 4
        self.header_info['missing_frames'] = unpack_from('H', self.master_header, 103)[0]
        self.header_info['battery_voltage_mV'] = unpack_from('H', self.master_header, 105)[0]
        self.header_info['min_signal'] = unpack_from('f', self.master_header, 107)[0]
        self.header_info['max_signal'] = unpack_from('f', self.master_header, 111)[0]


class DecimatedContinuousReader(_TSReaderBase):
    """
    Class to create a streamed reader for decimated continuous time series.
    Extension file: *.td_150, *.td_30.
    """
    def __init__(self, leap_seconds: int, header_size: int = 128):
        """Instantiate DecimatedContinuousReader class."""
        super().__init__(leap_seconds, header_size)

    def parse_data_file(self, data_file: Path) -> ndarray:
        """
        Parse decimated continous data file.
        Each file contains a master header (128 bytes long) and a decimated
        data segment.
        Decimated data are packed as a raw sequence of IEEE 754 single
        precision floating point (32 bits) samples, scaled to "Volts at
        instrument input" by default.
        """
        logger.info("Parsing data file - {}.".format(data_file))
        with open(data_file, 'rb') as stream:
            self.master_header = stream.read(self.header_size)
            self.unpack_header()
            # - Compare expected number of data points vs actual number of data points
            #   Expected number of data points = self.header_info['frag_period'] * self.header_info['sample_rate']
            #   Actual number of data points = data_pts.size
            #   Generate log error message if unexpected mistmatch
            # - Check number of saturated frames
            #   Generate log warning message if unexpected number
            # - Check number of missing frames
            #   Generate log warning message if unexpected number
            # - Check number of bytes per sample (Should be 4 for decimated time-series and 3 for undecimated ones)
            #   Generate log error message if unexpected number
            if not check_cha_id(self.header_info['ch_id'], data_file):
                return empty([0])
            if not check_sample_rate(self.header_info['sample_rate'], data_file):
                return empty([0])
            data_pts = fromfile(stream, dtype=float32, count=-1)
            return data_pts

    def parse_data_files(self, data_files: List[Path]) -> None:
        """
        Parse all decimated continuous data files.
        The data attribute is a list of one ndarray. Data samples read from
        each file are appended to the same array (continuous data).
        The time_stamps attribute is a list of one UTC time stamp corresponding
        to the data acquisition start time.
        """
        data = empty([0])
        for data_file in data_files:
            data_ = self.parse_data_file(data_file)
            data = append(data, data_)
        self.data.append(data)
        time_stamp = gps_to_utc(self.header_info['rec_id'], self.leap_seconds,
                                decimated_continuous=True)
        self.time_stamps.append(time_stamp)


class DecimatedSegmentedReader(_TSReaderBase):
    """
    Class to create a streamed reader for decimated segmented time series.
    Extension file: *.td_24k, *.td_2400.
    """
    def __init__(self, leap_seconds: int, header_size: int = 128):
        """Instantiate DecimatedSegmentedReader class."""
        super().__init__(leap_seconds, header_size)

    def parse_subheader(self) -> None:
        """
        Parse decimated segment header.
        This sub-header is only present in the Decimated Segmented Files and is
        32 bytes long.
        """
        self.subheader = {}
        subheader_bytes = self.stream.read(32)
        if len(subheader_bytes) == 32:
            self.subheader['time_stamp'] = unpack_from('I', subheader_bytes, 0)[0]
            self.subheader['num_samples'] = unpack_from('I', subheader_bytes, 4)[0]
            self.subheader['sat_count'] = unpack_from('H', subheader_bytes, 8)[0]
            self.subheader['miss_count'] = unpack_from('H', subheader_bytes, 10)[0]
            self.subheader['min_val'] = unpack_from('f', subheader_bytes, 12)[0]
            self.subheader['max_val'] = unpack_from('f', subheader_bytes, 16)[0]
            self.subheader['avg_val'] = unpack_from('f', subheader_bytes, 20)[0]

    def parse_data_segment(self) -> ndarray:
        """
        Parse data segment.
        Each data segment hould be preceded by a segment header which contains
        a time stamp for the first data point in the segment, and the number of
        samples in the segment.
        """
        data_segment = fromfile(self.stream, dtype=float32, count=self.subheader['num_samples'])
        return data_segment

    def parse_data_file(self, data_file: Path) -> None:
        """
        Parse decimated segmented data file.
        Each file contains a master header (128 bytes long), a segment header
        (or sub-header that is 32 bytes long) and a decimated data segment.
        Decimated data are packed as a raw sequence of IEEE 754 single
        precision floating point (32 bits) samples, scaled to "Volts at
        instrument input" by default.
        """
        logger.info("Parsing data file - {}.".format(data_file))
        parse = True
        with open(data_file, 'rb') as self.stream:
            self.master_header = self.stream.read(self.header_size)
            self.unpack_header()
            # - Check number of saturated frames
            #   Generate log warning message if unexpected number
            # - Check number of missing frames
            #   Generate log warning message if unexpected number
            # - Check number of bytes per sample (Should be 4 for decimated time-series and 3 for undecimated ones)
            #   Generate log error message if unexpected number
            if not check_cha_id(self.header_info['ch_id'], data_file):
                parse = False
            if not check_sample_rate(self.header_info['sample_rate'], data_file):
                parse = False
            while parse:
                self.parse_subheader()
                if not self.subheader:
                    break
                # - Compare expected number of samples vs actual number of samples
                #   Expected number of samples = sample rate * interleave length
                #   Actual number of samples =  data_segment.size
                #   Generate log error message if unexpected mistmatch
                # - Check number of data segments
                #   Expected number of data segments = int((file length + period) / (period + interleave length))
                #   File length (in s) is provided by the fragmentation period header (frag_period)
                #   Interleave length (data segment length) and period are defined under decimation field in recmeta
                #   Generate log error message if unexpected value (?)
                data_segment = self.parse_data_segment()
                self.data.append(data_segment)
                time_stamp = gps_to_utc(self.subheader['time_stamp'], self.leap_seconds)
                self.time_stamps.append(time_stamp)

    def parse_data_files(self, data_files: List[Path]) -> None:
        """
        Parse all decimated segmented data files.
        The data attribute is a list of X ndarrays. The number of ndarrays is
        equal to the number of files times the number of data segment per file.
        The time_stamps attribute is a list of X UTC time stamps corresponding
        to the time stamp of first data point in each data segment.
        """
        for data_file in data_files:
            self.parse_data_file(data_file)


class NativeReader(_TSReaderBase):
    """
    Class to create a streamed reader for native sampling rate time series
    (i.e., continuous and non-decimated time series).
    Extension file: *.bin.
    """
    def __init__(self, leap_seconds: int, header_size: int = 128,
                 scale_to: int = DataScaling.AD_input_volts,
                 channel_gain: float = 0.5, ad_plus_minus_range: float = 5.0):
        """Instantiate NativeReader class."""
        super().__init__(leap_seconds, header_size)

        self.next_frame = 0  # next frame parsed by the reader (track missing frames)
        self.data_scaling = scale_to
        self.total_circuitry_gain = channel_gain
        self.ad_plus_minus_range = ad_plus_minus_range
        self.footer_idx_samp_mask = int('0x0fffffff', 16)
        self.footer_sat_mask = int('0x70000000', 16)

    def set_scale_factor(self):
        """Set data scaling factor."""
        # Voltage range at the input of the board.
        if self.data_scaling == DataScaling.AD_in_ADunits:
            self._scale_factor = 256
        elif self.data_scaling == DataScaling.AD_input_volts:
            self._scale_factor = self.ad_plus_minus_range / (2 ** 31)
        elif self.data_scaling == DataScaling.instrument_input_volts:
            self.input_plusminus_range = self.ad_plus_minus_range / self.total_circuitry_gain
            self._scale_factor = self.input_plusminus_range / (2 ** 31)
        else:
            logger.error("Invalid data scaling requested!")
            sys.exit(1)

    def parse_data_frame(self, data_frame: bytes) -> ndarray:
        """
        Parse sampled date in given dataframe.
        The sampled data are 60 bytes long organized as 20 consecutive 3-byte
        sample values, signed, big endian
        """
        samples = empty([0])
        for ptr_sample in range(0, 60, 3):
            sample = unpack(">i", data_frame[ptr_sample:ptr_sample + 3] + b'\x00')[0]
            sample *= self._scale_factor
            samples = append(samples, sample)
        return samples

    def check_skipped_frames(self, footer: int) -> bool:
        """
        Check for skipped data frames due to transfer loss during transmission.
        Necessary step to define correct start time for each data frame.
        """
        self.frame_count = footer & self.footer_idx_samp_mask
        dif_count = self.frame_count - self.next_frame
        self.next_frame = self.frame_count + 1
        if dif_count != 0:
            logger.warning("Missing data frame(s) - Frame counter: {} "
                           "(differential count: {})! Updating start time of "
                           "next data frame accordingly."
                           .format(self.frame_count, dif_count))
            return True
        return False

    def check_analog_saturation(self, footer: int):
        """
        Check if, for given data frame, there was saturation detected in the
        analog front end of electric channels.
        If any of the bits of this field is non-zero, it indicates that the
        front end saturated at some point while acquiring data for this frame.
        """
        sat_count = (footer & self.footer_sat_mask) >> 24
        if sat_count:
            logger.warning("Saturation detected in analog front end of the "
                           "electric channels - Frame counter: {} (saturation "
                           "value: {})!".format(self.frame_count, sat_count))

    def get_time_stamp(self, start_time: UTCDateTime) -> UTCDateTime:
        """
        Get start time of given data frame.
        Each data frame contains 20 consecutive 3-byte sample values.
        """
        time_stamp = start_time + self.frame_count * 20 / self.header_info['sample_rate']
        return time_stamp

    def parse_data_file(self, data_file: Path) -> None:
        """
        Parse continuous data file.
        Each file contains a master header (128 bytes long) followed by a data
        payload. The data payload in each continuous file is arranged in
        consecutive blocks of 64 bytes (data frames). Each 64-byte data frame
        contains:
        - 60 bytes worth of sampled data, arranged as 20 consecutive 3-byte
          sample values, signed, big endian.
        - 4 bytes containing a frame footer, arranged as an unsigned little
          endian integer.
        Each sampled data point should be scaled to "Volts at instrument input".
        !!! The scale factor to go from counts to volts in currently not properly
        implemented. Units of data parsed by the Native Reader is currently counts !!!
        In the data frame footer: the most significant bit is a flag for
        internal usage; the 3 next most significant bits are used as a hardware
        saturation count for the frame; the remaining bits are used as a
        sequential, unsigned frame counter, which increases every frame (this
        counter is used to detect frame transfer loss during transmission).
        """
        logger.info("Parsing data file - {}.".format(data_file))
        parse = True
        with open(data_file, 'rb') as stream:
            self.master_header = stream.read(self.header_size)
            self.unpack_header()
            self.set_scale_factor()
            # - Check number of saturated frames
            #   Generate log warning message if unexpected number
            # - Check number of missing frames
            #   Generate log warning message if unexpected number
            # - Check number of bytes per sample (Should be 4 for decimated time-series and 3 for undecimated ones)
            #   Generate log error message if unexpected number
            if not check_cha_id(self.header_info['ch_id'], data_file):
                parse = False
            if not check_sample_rate(self.header_info['sample_rate'], data_file):
                parse = False
            if parse:
                data = empty([0])
                start_time = gps_to_utc(self.header_info['rec_id'], self.leap_seconds)
                while True:
                    data_frame = stream.read(64)
                    if len(data_frame) < 64:
                        break
                    samples = self.parse_data_frame(data_frame)
                    footer = unpack_from('I', data_frame, 60)[0]
                    skipped_frames = self.check_skipped_frames(footer)
                    self.check_analog_saturation(footer)
                    if skipped_frames:
                        if data.size != 0:
                            self.data.append(data)
                            data = empty([0])
                        data = append(data, samples)
                        self.time_stamps.append(self.get_time_stamp(start_time))
                    else:
                        if data.size == 0:
                            self.time_stamps.append(self.get_time_stamp(start_time))
                        data = append(data, samples)
                self.data.append(data)

    def parse_data_files(self, data_files: List[Path]) -> None:
        """
        Parse all native sampling rate data files.
        The data attribute is a list of X ndarrays. The number of ndarrays
        depends on whether there are any gaps in the continous data from data
        frame transfer loss.
        The time_stamps attribute is a list of X UTC time stamps corresponding
        to the time stamp of the first data point in each continuous data chunk.
        """
        for data_file in data_files:
            self.parse_data_file(data_file)
