# -*- coding: utf-8 -*-
"""
Routines that handle Phoenix instrument response.

-----

We are currently waiting for Phoenix Geophysics to provide us more information
about the different response stages of the Phoenix recording system.

The only stage characterized at this point is the analog sensor reponse (stage 1).
Until every other stages are fully characterized, we will use a "dummy" response to
build a Response object used in the metadata_conversion module to generate StationXML.

The generated StationXML will be flagged using the "Site Name" metadata field to
indicate to the end-use that the archived response is incorrect. This flag will
also allow to keep track of the StationXML that needs updating once the response of
the Phoenix recording system is fully characterized.

We will use a single stage response (coefficient multiplier filter type) to represent
the "dummy" filter.

-----

Maeva Pourpoint - IRIS/PASSCAL
"""

from __future__ import annotations

from obspy.core.inventory import Response, CoefficientsTypeResponseStage, InstrumentSensitivity

from phoenix2seed.logging import setup_logger

# Set up logging
logger = setup_logger(__name__)

CHA_STAGE_GAIN = 1.0
CHA_STAGE_GAIN_FREQUENCY = 0.0
CHA_UNITS = ['V', 'volts']


class PhoenixResponse:
    """
    This class currently handles the creation of a single stage response for each data
    channel ("dummy" response).
    """

    def __init__(self, sample_rate: int) -> None:
        """
        Instantiate PhoenixResponse class.
        """
        self.sample_rate = sample_rate

    def create_coefficient_filter(self) -> CoefficientsTypeResponseStage:
        """
        Create a coefficient multiplier filter (frequency independant gain).
        A coefficient multiplier filter can be represented as a FIR filter with a
        single coefficient (i.e. number of numerators is 1 and denominators is 0).
        When no denominators are provided, the filter must be digital.
        """
        cf_filter = CoefficientsTypeResponseStage(stage_sequence_number=1,
                                                  stage_gain=CHA_STAGE_GAIN,
                                                  stage_gain_frequency=CHA_STAGE_GAIN_FREQUENCY,
                                                  input_units=CHA_UNITS[0],
                                                  output_units=CHA_UNITS[0],
                                                  cf_transfer_function_type='DIGITAL',
                                                  decimation_input_sample_rate=self.sample_rate,
                                                  decimation_factor=1,
                                                  decimation_offset=0,
                                                  decimation_delay=0.0,
                                                  decimation_correction=0.0,
                                                  numerator=[1],
                                                  denominator=[],
                                                  description="coefficient multiplier filter",
                                                  input_units_description=CHA_UNITS[1],
                                                  output_units_description=CHA_UNITS[1])
        return cf_filter

    def get_cha_resp(self) -> Response:
        """
        Create a single stage response (CoefficientsTypeResponseStage) for
        all channels ("dummy" response).
        """
        cha_resp = Response()
        cha_resp.instrument_sensitivity = InstrumentSensitivity(value=CHA_STAGE_GAIN,
                                                                frequency=CHA_STAGE_GAIN_FREQUENCY,
                                                                input_units=CHA_UNITS[0],
                                                                output_units=CHA_UNITS[0],
                                                                input_units_description=CHA_UNITS[1],
                                                                output_units_description=CHA_UNITS[1])
        cf_filter = self.create_coefficient_filter()
        cha_resp.response_stages = [cf_filter]
        return cha_resp
