# -*- coding: utf-8 -*-
"""
Various "utils" functions required for data and metadata formatting and
validation.

Maeva Pourpoint - IRIS/PASSCAL
"""

from __future__ import annotations

import re

from collections import defaultdict
from functools import singledispatch
from obspy import UTCDateTime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from phoenix2seed.logging import setup_logger

# Set up logging
logger = setup_logger(__name__)

ELEC_DEFAULT = 'Borin STELTH 4 - Silver-Silver Chloride'
COIL_DEFAULT = {'MTC-150': {'manufacturer': 'Phoenix Geophysics',
                            'model': 'MTC-150',
                            'type': 'ultra wideband'},
                'MTC-180': {'manufacturer': 'Phoenix Geophysics',
                            'model': 'MTC-180',
                            'type': 'ultra wideband'},
                'MTC-155': {'manufacturer': 'Phoenix Geophysics',
                            'model': 'MTC-155',
                            'type': 'ultra wideband'},
                'MTC-185': {'manufacturer': 'Phoenix Geophysics',
                            'model': 'MTC-185',
                            'type': 'ultra wideband'},
                'ANT/4': {'manufacturer': 'Zonge International',
                          'model': 'ANT/4',
                          'type': 'wideband'},
                'ANT/7': {'manufacturer': 'Zonge International',
                          'model': 'ANT/7',
                          'type': 'wideband'}}

# Maximum number of electrode pairs or coils than can be connected to an MTU-5C
# at a time
NUM_CHA_MAX = {'elec': 2, 'mag': 3}

# Valid file extension and associated: sampling rate (Hz) and reader
VALID_EXTENSION = {'bin': ['24000', 'NativeReader'],
                   'td_30': ['30', 'DecimatedContinuousReader'],
                   'td_150': ['150', 'DecimatedContinuousReader'],
                   'td_2400': ['2400', 'DecimatedSegmentedReader'],
                   'td_24k': ['24000', 'DecimatedSegmentedReader']}


@singledispatch
def check_cha_id(arg1, arg2):
    """
    Check channel id.
    Expected type for arg1: str or int.
    Expected type for arg2: Path.
    """
    logger.warning("Unsupported type for channel id ({}). The file master "
                   "header may have been corrupted! Skipping file {}."
                   .format(arg1, arg2))
    return False


@check_cha_id.register
def _(arg1: str, arg2: Path) -> bool:
    """Check channel id in folder name against channel id in file name."""
    cha_id_folder = arg2.parts[-1]
    cha_id_file = arg1.split('_')[2]
    if cha_id_folder != cha_id_file:
        file = arg2.joinpath(arg1)
        logger.warning("Mismatch between channel ids recorded in folder name "
                       "vs file name! Skipping file {}.".format(file))
        return False
    return True


@check_cha_id.register
def _(arg1: int, arg2: Path) -> bool:
    """
    Check channel id in folder name against channel id recorded in file master
    header.
    """
    cha_id_folder = arg2.parts[-2]
    cha_id_header = str(arg1)
    if cha_id_folder != cha_id_header:
        logger.warning("Mismatch between channel ids recorded in folder name "
                       "and the file master header! Skipping file {}."
                       .format(arg2))
        return False
    return True


def check_email_format(email: str) -> bool:
    """
    Basic email check. Check for proper formatting.
    Will not catch typos or fake email addresses.
    """
    email_pattern = r"[\w\.\-_]+@[\w\.\-_]+"
    try:
        valid = re.match(email_pattern, email)
    except TypeError:
        logger.error("The provided email '{}' should be a string.".format(email))
        return False
    else:
        if valid is None:
            logger.error("Invalid email. The provided email '{}' does not "
                         "meet minimum formatting requirements: "
                         "account@domain.".format(email))
        return bool(valid)


def check_inst_specs(specs: str, equipment: str, cha_info: str = "") -> bool:
    """
    Check instrument specifications provided by the user.
    Both manufacturer and model info are required inputs. Type is optional.
    """
    if specs in [ELEC_DEFAULT, *COIL_DEFAULT]:
        return True
    elif specs is not None and re.match(r"^Manufacturer: \w* - Model: \w* - Type: \w*$", specs):
        manufacturer, model, _ = [x.split(':')[-1].strip() for x in specs.split('-')]
        if manufacturer == '':
            logger.error("Please provide {} manufacturer {}. Required "
                         "metadata field.".format(equipment, cha_info))
            return False
        if model == '':
            logger.error("Please provide {} model {}. Required metadata "
                         "field.".format(equipment, cha_info))
            return False
        return True
    else:
        logger.error("Please provide {} specs {}. Required metadata "
                     "field.".format(equipment, cha_info))
        return False


def check_sample_rate(rate_header: int, file: Path) -> bool:
    """
    Check sample rate in file extension against sample recorded in file master
    header.
    """
    rate_extension = int(VALID_EXTENSION.get(file.suffix[1:], ['0'])[0])
    if rate_header != rate_extension:
        logger.warning("Mismatch between sample rate inferred from file "
                       "name and the sample rate recorded in the file "
                       "master header! Skipping file {}.".format(file))
        return False
    return True


def check_sn(equipment: str, sn: str, data_input: Optional[str] = None) -> bool:
    """
    Check user provided serial number for a data logger or coil against
    the serial number recorded in the recmeta.json file.
    The serial numbers of the data logger and coils installed at a given station
    will only be recorded in the recmeta.json file if they were programmed in
    using EMpower.
    """
    if data_input is not None:
        valid = (sn == data_input)
        if not valid:
            logger.error("The serial number of the {} differs from the one "
                         "programmed with EMpower in the field ({} != {}). "
                         "Please correct!".format(equipment, sn, data_input))
    else:
        if "coil" in equipment:
            sn_len = "{3,5}"
            msg = ("The serial number of the {} should be between 3 and 5 "
                   "digits long (ANT/7: 3 digits long, ANT/4: 4 digits long, "
                   "MTC-1**: 5 digits long).".format(equipment))
        else:
            sn_len = "{5}"
            msg = ("The serial number of the {} should be 5 digits long."
                   .format(equipment))
        valid = re.match(rf"^\d{sn_len}$", sn)  # type: ignore
        valid = False if valid is None else True
        if not valid:
            logger.error(msg)
    return valid


def get_ids(num_runs: int, cha_type: str) -> List[Tuple[str, int]]:
    """Get permutation of run ids and channel numbers."""
    run_list = get_run_list(num_runs)
    return [(i, j) for i in run_list for j in range(1, NUM_CHA_MAX[cha_type] + 1)]


def get_run_list(num_runs: int) -> List[str]:
    """Get list of run ids."""
    return [chr(x) for x in range(ord('a'), ord('a') + num_runs)]


def gps_to_utc(unix_gps: int, leap_seconds: int, decimated_continuous: bool = False) -> UTCDateTime:
    """
    Convert a Unix-epoch based timestamp from GPS to UTC and reformat as
    %Y-%m-%d %H:%M:%S.
    For decimated continous data, the start of the data contained is one
    second after the start of the recording given by the Master Header of
    the file.
    """
    unix_utc = unix_gps - leap_seconds
    if decimated_continuous:
        unix_utc += 1
    return UTCDateTime(unix_utc)


def is_empty(val: Any) -> bool:
    """Check that specific data structure is empty."""
    if val is None:
        return True
    if (isinstance(val, list) or isinstance(val, dict)) and not val:
        return True
    return False


def is_valid_uri(uri: str) -> bool:
    """From obpsy.core.inventory.util private function _is_valid_uri"""
    if ':' not in uri:
        return False
    scheme, path = uri.split(':', 1)
    if any(not x.strip() for x in (scheme, path)):
        return False
    return True


def merge_nested_dict(dict1: Dict, dict2: Dict) -> Dict:
    """Merge two nested dictionaries."""
    merged_dict: Dict = defaultdict(dict)
    merged_dict.update(dict1)
    for key, val in dict2.items():
        merged_dict[key].update(val)
    return dict(merged_dict)


def str2list(str_input: str, sep: Optional[str] = ',') -> List[str]:
    """Convert comma separated string input to a list of strings."""
    return [x.strip() for x in str_input.split(sep) if x != '']
